using System;
using System.Collections.Generic;
using System.Text;
using C1.C1Report;

namespace ReportDesignerApp
{
    public class EditorScriptItem
    {
        public static class Names
        {
            public const string Global = "GlobalScripts";
            public const string Report = "Report";
            public const string DataSource = "Report.DataSource";
            public const string UnnamedSection = "<section>";
            public const string UnnamedField = "<field>";
            public const string NewField = "<New Calculated Field>";
            public const string StandaloneExpression = "Expression";
        }

        /// <summary>
        /// The kind of script item.
        /// </summary>
        public enum Kind
        {
            /// <summary>
            /// Report-scope script (includes the global script).
            /// </summary>
            ReportScript,
            /// <summary>
            /// The DataSource.Filter "expression". This is a very special case, not VBScript at all.
            /// </summary>
            DataSourceFilter,
            /// <summary>
            /// Section-scope script.
            /// </summary>
            SectionScript,
            /// <summary>
            /// Text of a Calculated field.
            /// </summary>
            FieldExpression,
            /// <summary>
            /// Text of a non-calculated field.
            /// </summary>
            FieldText,
            /// <summary>
            /// Expression for a new (not yet added) field.
            /// </summary>
            NewFieldExpression,
            /// <summary>
            /// Expression with unknown source (e.g. from Maps custom field).
            /// </summary>
            StandaloneExpression,
        }
        public EditorScriptItem(Kind itemKind, int index, string hostName, string name, string text)
        {
            ItemKind = itemKind;
            Index = index;
            HostName = hostName;
            Name = name;
            Text = text;
        }

        public static EditorScriptItem FromEditContext(C1Report report)
        {
            if (report == null)
                return null;

            System.Diagnostics.Debug.Assert(report.EditContext != null);
            if (report.EditContext == null)
                return null;

            EditorScriptItem si = null;
            var pd = report.EditContext.PropertyDescriptor;
            var obj = report.EditContext.Instance;
            if (pd == null || obj == null)
                return null;

            var value = pd.GetValue(obj) as string;
            if (obj is C1Report)
            {
                si = new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, pd.Name, value);
            }
            else if (obj is Section)
            {
                Section s = obj as Section;
                si = new EditorScriptItem(EditorScriptItem.Kind.SectionScript, (int)s.Type, s.Name, pd.Name, value);
            }
            else if (obj is Field)
            {
                Field f = obj as Field;

                bool calculated = f.Calculated; // (bool)typeof(Field).GetProperty("Calculated").GetValue(obj);
                string fName = f.Name;
                Section s = report.Sections[f.Section];
                string sName = s != null ? s.Name : string.Empty;

                si = new EditorScriptItem(calculated ? EditorScriptItem.Kind.FieldExpression : EditorScriptItem.Kind.FieldText, f.Index, s.Name, fName, value);
            }
            else if (obj is C1.C1Report.DataSource)
            {
                si = new EditorScriptItem(EditorScriptItem.Kind.DataSourceFilter, -1, EditorScriptItem.Names.DataSource, pd.Name, value);
            }
            else // unknown expression (e.g. called from Maps)
            {
                // Note: as we are not sure what the object is, we try its ToString but try to keep it sane:
                var objName = obj.ToString().Trim();
                if (objName.Contains(" ") || objName.Length > 64)
                    objName = EditorScriptItem.Names.StandaloneExpression;

                si = new EditorScriptItem(EditorScriptItem.Kind.StandaloneExpression, -1, objName, pd.Name, value);
            }
            return si;
        }

        public static EditorScriptItem FromNewField(C1Report report, string text)
        {
            if (report == null)
                return null;

            EditorScriptItem si = new EditorScriptItem(EditorScriptItem.Kind.NewFieldExpression, -1, EditorScriptItem.Names.NewField, "Text", text);
            return si;
        }

        public static EditorScriptItem FromGlobalScripts(C1Report report)
        {
            if (report == null)
                return null;

            EditorScriptItem si = new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, EditorScriptItem.Names.Global, string.Empty);
            return si;
        }

        public Kind ItemKind { get; private set; }
        /// <summary>
        /// For fields and sections, the index in the collection (needed because fields/sections' names may be duplicate or empty).
        /// For other objects, use -1.
        /// </summary>
        public int Index { get; private set; }
        public string HostName { get; private set; }
        public string Name { get; private set; }
        public string Text { get; private set; }

        public bool HasText { get { return !string.IsNullOrEmpty(Text); } }

        public bool IsFieldExpression { get { return ItemKind == Kind.FieldExpression || ItemKind == Kind.FieldText || ItemKind == Kind.NewFieldExpression; } }

        public bool CanSyntaxCheck(out bool asExpression)
        {
            switch (ItemKind)
            {
                case Kind.ReportScript:
                case Kind.SectionScript:
                    asExpression = false;
                    return true;
                case Kind.FieldExpression:
                case Kind.FieldText:
                case Kind.NewFieldExpression:
                    asExpression = true;
                    return true;
                case Kind.DataSourceFilter:
                case Kind.StandaloneExpression:
                    asExpression = false;
                    return false;
                default:
                    System.Diagnostics.Debug.Assert(false);
                    asExpression = false;
                    return false;
            }
        }

        /// <summary>
        /// Fires when a script item is written back.
        /// Useful for setting the dirty flag on the report being edited.
        /// </summary>
        public static EventHandler Written;

        /// <summary>
        /// Keys are used to identify script items.
        /// Additionally, items are sorted by key in UI.
        /// </summary>
        public string Key
        {
            get
            {
                string prefix = string.Empty;
                switch (ItemKind)
                {
                    case Kind.ReportScript:
                        prefix = IsGlobal ? "1." : "2.";
                        break;
                    case Kind.DataSourceFilter:
                        prefix = "3.";
                        break;
                    case Kind.SectionScript:
                        prefix = "4.";
                        break;
                    case Kind.FieldExpression:
                        prefix = "5.";
                        break;
                    case Kind.FieldText:
                        prefix = "5."; // use "6." to put non-calculated fields after calculated
                        break;
                    case Kind.NewFieldExpression:
                    case Kind.StandaloneExpression:
                        prefix = "0.";
                        break;
                    default:
                        System.Diagnostics.Debug.Assert(false);
                        break;
                }
                return string.Format("{0}.{1}.{2}.{3}", prefix, HostName, Index, Name);
            }
        }

        public override string ToString()
        {
            return string.Format("{0}.{1}", HostName, Name);
        }

        public bool IsGlobal
        {
            get { return HostName == Names.Report && Name == Names.Global; }
        }

        public void WriteValue(C1Report report, string value)
        {
            if (report == null)
                return;

            Text = value;
            switch (ItemKind)
            {
                case Kind.ReportScript:
                    typeof(C1Report).GetProperty(Name).SetValue(report, Text, null);
                    break;
                case Kind.DataSourceFilter:
                    System.Diagnostics.Debug.Assert(HostName == Names.DataSource);
                    typeof(C1.C1Report.DataSource).GetProperty(Name).SetValue(report.DataSource, Text, null);
                    break;
                case Kind.SectionScript:
                    typeof(Section).GetProperty(Name).SetValue(report.Sections[HostName], Text, null);
                    break;
                case Kind.FieldExpression:
                case Kind.FieldText:
                    typeof(Field).GetProperty("Text").SetValue(report.Fields[Name], Text, null);
                    break;
                case Kind.NewFieldExpression: // saved only in item's text.
                case Kind.StandaloneExpression:
                    break;
                default:
                    System.Diagnostics.Debug.Assert(false);
                    break;
            }
            if (Written != null)
                Written(this, EventArgs.Empty);
        }
    }

    public class EditorScriptItemCollection : SortedDictionary<string, EditorScriptItem>
    {
        public EditorScriptItemCollection(C1Report report)
        {
            Action<EditorScriptItem> addScriptItem = (si_) => this.Add(si_.Key, si_);

            // global script:
            addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, EditorScriptItem.Names.Global, report.GlobalScripts));

            // report event handlers:
            addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, "OnOpen", report.OnOpen));
            addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, "OnClose", report.OnClose));
            addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, "OnNoData", report.OnNoData));
            addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, "OnPage", report.OnPage));
            addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.ReportScript, -1, EditorScriptItem.Names.Report, "OnError", report.OnError));

            // data filter:
            addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.DataSourceFilter, -1, EditorScriptItem.Names.DataSource, "Filter", report.DataSource.Filter));

            // Sections' event handlers:
            for (int i = 0; i < report.Sections.Count; ++i)
            {
                var s = report.Sections[i];
                var index = (int)s.Type;
                var name = string.IsNullOrEmpty(s.Name) ? EditorScriptItem.Names.UnnamedSection : s.Name;
                addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.SectionScript, index, name, "OnFormat", s.OnFormat));
                addScriptItem(new EditorScriptItem(EditorScriptItem.Kind.SectionScript, index, name, "OnPrint", s.OnPrint));
            }

            // Report fields' expressions:
            for (int i = 0; i < report.Fields.Count; ++i)
            {
                var f = report.Fields[i];
                var s = report.Sections[f.Section];
                string sName = s != null ? s.Name : string.Empty;
                var fName = string.IsNullOrEmpty(f.Name) ? EditorScriptItem.Names.UnnamedField : f.Name;
                addScriptItem(new EditorScriptItem(f.Calculated ? EditorScriptItem.Kind.FieldExpression : EditorScriptItem.Kind.FieldText, f.Index, sName, fName, f.Text));
            }
        }

        public EventHandler ItemWritten;
    }
}
