//----------------------------------------------------------------------------
// FieldWrappers.cs
//
// Wrapper classes for C1Report.Field objects. These classes expose only a
// subset of the many field properties, making the designer's property window
// easier to manage.
//
// This is a part of the C1ReportDesigner package.
// Copyright (C) GrapeCity, Inc
// All rights reserved.
//
// This source code is intended as a supplement to the
// C1ReportDesigner control and related documentation 
// provided with the product.
// See these sources for detailed information regarding the
// C1ReportDesigner product.
//----------------------------------------------------------------------------
using System;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Design;
using System.ComponentModel;

using C1.C1Report;
using C1.C1Report.Util;
using C1.C1Report.Design;

namespace ReportDesignerApp
{
	/// <summary>
	/// FieldWrapper.
	/// Base field wrapper class, contains properties that are 
	/// common to all fields, all wrappers derive from this.
	/// </summary>
	public class FieldWrapper
	{
		// ** fields
		protected Field _field;

		// ** ctor
		public FieldWrapper(Field field)
		{
			_field = field;
		}

		// ** static FieldWrapper factories
		static public object Create(Field field)
		{
			// subreport fields
			if (field.Subreport != null)
			{
				return new SubreportField(field);
			}

			// special fields
			if (field.RTF)
			{
				return new RtfField(field);
			}
			if (field.CheckBox != CheckBoxEnum.NoCheckBox)
			{
				return new CheckBoxField(field);
			}
			if (field.BarCode != BarCodeEnum.None)
			{
				return new BarCodeField(field);
			}

			// image fields
			if (field.Picture != null)
			{
				return (field.Calculated)
					? (FieldWrapper)new BoundImageField(field)
					: (FieldWrapper)new UnboundImageField(field);
			}

            // shape fields
            if (field.Shape != null)
                return new ShapeField(field);

			// label/calculated fields
			if (field.Text != null && field.Text.Length > 0)
			{
				return (field.Calculated)
					? new CalculatedField(field)
					: new LabelField(field);
			}

			// box fields
			if (field.BorderStyle != BorderStyleEnum.Transparent ||
				field.BackColor   != Color.Transparent)
			{
				return new BoxField(field);
			}

			// page break field
			if (field.ForcePageBreak != ForcePageBreakEnum.None)
			{
				return new PageBreakField(field);
			}

			// this should be a custom field, otherwise it seems useless...
			return field;
		}

		static public object[] Create(object[] selection)
		{
			object[] wrapper = new object[selection.Length];
			for (int i = 0; i < wrapper.Length; i++)
			{
				wrapper[i] = Create((Field)selection[i]);
			}
			return wrapper;
		}

		// ** implicit conversion to field
		public static implicit operator Field(FieldWrapper wrapper)
		{
			return wrapper._field;
		}
		
		// ** elements common to all field types

		// for use by designer
		[Browsable(false)]
		public C1Report ParentReport
		{
			get { return _field.ParentReport; }
		}

		#region ** design
        [Category("Design")]
		[ParenthesizePropertyName(true)]
		[MergableProperty(false)]
		[DefaultValue(null)]
		public string Name
		{
			get { return _field.Name; }
			set { _field.Name = value; }
		}

        [Category("Design")]
        [DefaultValue(true)]
		public bool Visible
		{
			get { return _field.Visible; }
			set { _field.Visible = value; }
		}

        [Category("Design")]
        [TypeConverter(typeof(StringConverter))]
		[DefaultValue(null)]
		public object Tag
		{
			get { return _field.Tag; }
			set { _field.Tag = value; }
		}
		#endregion

		#region ** layout
        [Category("Layout")]
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public FieldBounds Bounds
		{
			get 
			{
				if (_bounds == null)
					_bounds = new FieldBounds(_field); 
				return _bounds;
			}
		}

		private FieldBounds _bounds;
        [Category("Layout")]
        [DefaultValue(ForcePageBreakEnum.None)]
		public ForcePageBreakEnum ForcePageBreak
		{
			get { return _field.ForcePageBreak; }
			set { _field.ForcePageBreak = value; }
		}
		#endregion
	}

	/// <summary>
	/// PageBreakField
	/// </summary>
	public class PageBreakField : FieldWrapper
	{
		public PageBreakField(Field field) : base(field) {}
	}

	
	/// <summary>
	/// AreaField
	/// Wraps fields that render into a rectangular area on the report.
	/// Most fields do, so this is a base class for most fields
	/// (PageBreak fields are the exception).
	/// </summary>
	public abstract class AreaField : FieldWrapper
	{
		public AreaField(Field field) : base(field) {}

		#region ** layout
        [Category("Layout")]
        [DefaultValue(AnchorEnum.Top)]
		public AnchorEnum Anchor
		{
			get { return _field.Anchor; }
			set { _field.Anchor = value; }
		}
		#endregion

		#region ** border
        [Category("Border")]
		[DefaultValue(BorderStyleEnum.Transparent)]
		public BorderStyleEnum BorderStyle
		{
			get { return _field.BorderStyle; }
			set { _field.BorderStyle = value; }
		}

        [Category("Border")]
        [DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double LineWidth
		{
			get { return _field.LineWidth; }
			set { _field.LineWidth = value; }
		}

        [Category("Border")]
        [DefaultValue(typeof(Color), "Black")]
		public Color BorderColor
		{
			get { return _field.BorderColor; }
			set { _field.BorderColor = value; }
		}
		#endregion
	}


	/// <summary>
	/// SolidAreaField
	/// AreaField plus BackColor/CanGrow/CanShrink
	/// </summary>
	public abstract class SolidAreaField : AreaField
	{
		public SolidAreaField(Field field) : base(field) {}

		#region ** layout
        [Category("Layout")]
		[DefaultValue(false)]
		public virtual bool CanGrow
		{
			get { return _field.CanGrow; }
			set { _field.CanGrow = value; }
		}

        [Category("Layout")]
        [DefaultValue(false)]
		public virtual bool CanShrink
		{
			get { return _field.CanShrink; }
			set { _field.CanShrink = value; }
		}
		#endregion

		#region ** appearance
        [Category("Appearance")]
        [DefaultValue(typeof(Color), "Transparent")]
		public virtual Color BackColor
		{
			get { return _field.BackColor; }
			set { _field.BackColor = value; }
		}
		#endregion
	}

	/// <summary>
	/// BoxField
	/// AreaField plus BackColor
	/// </summary>
	public class BoxField : AreaField
	{
		public BoxField(Field field) : base(field) {}

		#region ** appearance
        [Category("Appearance")]
		[DefaultValue(typeof(Color), "Transparent")]
		public virtual Color BackColor
		{
			get { return _field.BackColor; }
			set { _field.BackColor = value; }
		}
		#endregion
	}

	/// <summary>
	/// LineField
	/// </summary>
	public class LineField : AreaField
	{
		public LineField(Field field) : base(field) {}

		#region ** border
        [Category("Border")]
		[DefaultValue(LineSlantEnum.NoLine)]
#pragma warning disable 618
		public LineSlantEnum LineSlant
		{
			get { return _field.LineSlant; }
			set 
			{
				if (value != LineSlantEnum.NoLine)
					_field.LineSlant = value;
			}
		}
#pragma warning restore 618
		#endregion
	}

    /// <summary>
    /// Shape field.
    /// </summary>
    public class ShapeField : AreaField
    {
        #region Constructors
        public ShapeField(Field field)
            : base(field)
        {
        }
        #endregion

        #region Public properties
        [Category("Appearance")]
        [DefaultValue(ShapeType.None)]
        public ShapeType ShapeType
        {
            get { return _field.ShapeType; }
            set { _field.ShapeType = value; }
        }

        [Category("Appearance")]
        [TypeConverter(typeof(ExpandableObjectConverter))]
        public ShapeBase Shape
        {
            get { return _field.Shape; }
        }
        #endregion
    }


	/// <summary>
	/// BoundImageField
	/// </summary>
	public class BoundImageField : SolidAreaField
	{
		public BoundImageField(Field field) : base(field) {}

		#region ** appearance
        [Category("Appearance")]
		[DefaultValue(null)]
		[Editor(typeof(PictureHolderEditor), typeof(UITypeEditor))]
		public object Picture
		{
			get { return _field.Picture; }
			set { _field.Picture = value; }
		}

        [Category("Appearance")]
		[DefaultValue(PictureAlignEnum.LeftTop)]
		[Editor(typeof(PictureAlignEditor), typeof(UITypeEditor))]
		public PictureAlignEnum PictureAlign
		{
			get { return _field.PictureAlign; }
			set { _field.PictureAlign = value; }
		}

        [Category("Appearance")]
		[DefaultValue(PictureScaleEnum.Clip)]
		public PictureScaleEnum PictureScale
		{
			get { return _field.PictureScale; }
			set { _field.PictureScale = value; }
		}
		#endregion
	}


	/// <summary>
	/// UnboundImageField
	/// </summary>
	public class UnboundImageField : SolidAreaField
	{
		public UnboundImageField(Field field) : base(field) {}

		#region ** appearance
        [Category("Appearance")]
		[DefaultValue(null)]
		[Editor(typeof(ImageEditor), typeof(UITypeEditor))]
		public object Picture
		{
			get { return _field.Picture; }
			set { _field.Picture = value; }
		}

        [Category("Appearance")]
		[DefaultValue(PictureAlignEnum.LeftTop)]
		[Editor(typeof(PictureAlignEditor), typeof(UITypeEditor))]
		public PictureAlignEnum PictureAlign
		{
			get { return _field.PictureAlign; }
			set { _field.PictureAlign = value; }
		}

        [Category("Appearance")]
		[DefaultValue(PictureScaleEnum.Clip)]
		public PictureScaleEnum PictureScale
		{
			get { return _field.PictureScale; }
			set { _field.PictureScale = value; }
		}
		#endregion
	}


	/// <summary>
	/// SubreportField
	/// </summary>
	public class SubreportField : SolidAreaField
	{
		public SubreportField(Field field) : base(field)
		{
			_field.Calculated = true;
		}

		#region ** subreport
        [Category("Subreport")]
        [Description("Specifies the subreport to render into the field.")]
		[TypeConverter(typeof(SubreportFieldConverter))]
		[DefaultValue(null)]
		public C1Report Subreport
		{
			get { return _field.Subreport; }
			set { _field.Subreport = value; }
		}

        [Category("Subreport")]
        [Description("Specifies a filter expression used to link fields in the main and subreports.")]
        [Editor(typeof(TextFieldEditor), typeof(UITypeEditor))]
		[DefaultValue(null)]
		public string Text
		{
			get { return _field.Text; }
			set { _field.Text = value; }
		}
		#endregion
	}


	/// <summary>
	/// LabelField
	/// regular text field (not calculated)
	/// </summary>
	public class LabelField : SolidAreaField
	{
		public LabelField(Field field) : base(field) {}

		#region ** layout
        [Category("Layout")]
		[DefaultValue(false)]
		public virtual bool KeepTogether
		{
			get { return _field.KeepTogether; }
			set { _field.KeepTogether = value; }
		}
		#endregion

        #region ** data
        [Category("Data")]
        [DefaultValue(false)]
        public virtual bool Calculated
        {
            get { return _field.Calculated; }
            set { _field.Calculated = value; }
        }
        #endregion
        
        #region ** appearance
        [Category("Appearance")]
		[DefaultValue(null)]
		public virtual string Text
		{
			get { return _field.Text; }
			set { _field.Text = value; }
		}

        [Category("Appearance")]
		public virtual FontHolder Font
		{
			get { return _field.Font; }
			set { _field.Font = value; }
		}

        [Category("Appearance")]
		[DefaultValue(typeof(Color), "Black")]
		public Color ForeColor
		{
			get { return _field.ForeColor; }
			set { _field.ForeColor = value; }
		}

        [Category("Appearance")]
		[DefaultValue(true)]
		public virtual bool WordWrap
		{
			get { return _field.WordWrap; }
			set { _field.WordWrap = value; }
		}

        [Category("Appearance")]
		[DefaultValue(TextDirectionEnum.Normal)]
		public virtual TextDirectionEnum TextDirection
		{
#pragma warning disable 618
            get { return _field.TextDirection; }
			set { _field.TextDirection = value; }
#pragma warning restore 618
        }

        [Category("Appearance")]
        [DefaultValue(0f)]
        public float TextAngle
        {
            get { return _field.TextAngle; }
            set { _field.TextAngle = value; }
        }

        [Category("Appearance")]
		[DefaultValue(FieldAlignEnum.General)]
		[Editor(typeof(TextAlignEditor), typeof(UITypeEditor))]
		public FieldAlignEnum Align
		{
			get { return _field.Align; }
			set { _field.Align = value; }
		}

        [Category("Appearance")]
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public FieldMargins Margins
		{
			get 
			{
				if (_margins == null)
					_margins = new FieldMargins(_field); 
				return _margins;
			}
		}
		private FieldMargins _margins;
		#endregion
	}


	/// <summary>
	/// CalculatedField
	/// </summary>
	public class CalculatedField : LabelField
	{
		public CalculatedField(Field field) : base(field)
		{
			_field.Calculated = true;
		}

		#region ** data
        [Category("Data")]
        [DefaultValue(true)] // << override to change default value
        override public bool Calculated
        {
            get { return _field.Calculated; }
            set { _field.Calculated = value; }
        }

        [Category("Data")]
		[Editor(typeof(TextFieldEditor), typeof(UITypeEditor))] // << override to change editor
		[DefaultValue(null)]
		override public string Text
		{
			get { return _field.Text; }
			set { _field.Text = value; }
		}
		#endregion

		#region ** appearance
        [Category("Appearance")]
		[DefaultValue(null)]
		[TypeConverter(typeof(FormatConverter))]
		public string Format
		{
			get { return _field.Format; }
			set { _field.Format = value; }
		}
		#endregion
	}


	/// <summary>
	/// RtfField
	/// </summary>
	public class RtfField : CalculatedField
	{
		public RtfField(Field field) : base(field)
		{
			_field.RTF = true;
		}

		#region ** hidden
		[Browsable(false)]
		override public TextDirectionEnum TextDirection
		{
			get { return TextDirectionEnum.Normal; }
			set {}
		}
		#endregion
	}

	
	/// <summary>
	/// NoFontField
	/// Base class for barcode, checkboxes: hides font/wordwrap/keeptogether/textdirection
	/// </summary>
	public abstract class NoFontField : CalculatedField
	{
		public NoFontField(Field field) : base(field) {}

		#region ** hidden
		[Browsable(false)]
		override public FontHolder Font
		{
			get { return _field.Font; }
			set {}
		}

        [Browsable(false)]
		override public bool WordWrap
		{
			get { return false; }
			set {}
		}

        [Browsable(false)]
		override public bool KeepTogether
		{
			get { return false; }
			set {}
		}

        [Browsable(false)]
		override public TextDirectionEnum TextDirection
		{
			get { return TextDirectionEnum.Normal; }
			set {}
		}
		#endregion
	}


	/// <summary>
	/// CheckBoxField
	/// </summary>
	public class CheckBoxField : NoFontField
	{
		public CheckBoxField(Field field) : base(field) {}

        [Category("Appearance")]
		[DefaultValue(CheckBoxEnum.CheckBox)]
		public CheckBoxEnum CheckBox
		{
			get { return _field.CheckBox; }
			set
			{
				if (value != CheckBoxEnum.NoCheckBox)
					_field.CheckBox = value; 
			}
		}
	}


	/// <summary>
	/// BarCodeField
	/// </summary>
	public class BarCodeField : NoFontField
	{
		public BarCodeField(Field field) : base(field) {}

		#region ** appearance
        [Category("Appearance")]
		[DefaultValue(BarCodeEnum.None)]
		public BarCodeEnum BarCode
		{
			get { return _field.BarCode; }
			set
			{
				if (value != BarCodeEnum.None)
					_field.BarCode = value; 
			}
		}
		#endregion
	}


	/// <summary>
	/// FieldMargins
	/// helper class to handle field margins
	/// </summary>
	public class FieldMargins
	{
		// ** fields
		Field _field;

		// ** ctor
		public FieldMargins(Field field)
		{
			_field = field;
		}

		// ** overrides
		public override string ToString()
		{
			return string.Format("{0}, {1}, {2}, {3}", Left, Top, Right, Bottom);
		}

		// ** object model
		[DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Left
		{
			get { return _field.MarginLeft; }
			set { _field.MarginLeft = value; }
		}

        [DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Top
		{
			get { return _field.MarginTop; }
			set { _field.MarginTop = value; }
		}

        [DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Right
		{
			get { return _field.MarginRight; }
			set { _field.MarginRight = value; }
		}

        [DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Bottom
		{
			get { return _field.MarginBottom; }
			set { _field.MarginBottom = value; }
		}
	}

	/// <summary>
	/// FieldBounds
	/// helper class to handle field bounds
	/// </summary>
	public class FieldBounds
	{
		// ** fields
		Field _field;

		// ** ctor
		public FieldBounds(Field field)
		{
			_field = field;
		}

		// ** overrides
		public override string ToString()
		{
			return string.Format("{0}, {1}, {2}, {3}", Left, Top, Width, Height);
		}

		// ** object model
		[DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Left
		{
			get { return _field.Left; }
			set { _field.Left = value; }
		}

        [DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Top
		{
			get { return _field.Top; }
			set { _field.Top = value; }
		}

        [DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Width
		{
			get { return _field.Width; }
			set { _field.Width = value; }
		}

        [DefaultValue(0.0)]
		[TypeConverter(typeof(PageUnitConverter))]
		public double Height
		{
			get { return _field.Height; }
			set { _field.Height = value; }
		}
	}
}
