using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;
#if CLR40 || CLR45
using System.Linq;
#endif

using C1.Win.C1Command;
using C1.C1Preview;
using C1.C1Report;
using C1.Win.C1InputPanel;
using ReportDesignerApp.Util;


namespace ReportDesignerApp.Forms
{
    public partial class ScriptEditorControl : UserControl
    {
        private C1Report _report; // report that owns the script
        private string[] _dbFields = null; // cached list of db fields, so that we can turn syntax hilite on/off.
        // private bool _syntaxHighlighting = true; // color text
        private SyntaxHighlightingType _syntaxHighlighting = SyntaxHighlightingType.None;
        private bool _showCalculatedFieldsOnly = false;
        private string _currentScriptItemKey = null;
        private bool _currentScriptIsDirty = false;
        private bool _currentScriptChangeInternal = false;
        EditorScriptItemCollection _scriptItems = null;
        // fonts used to hi-light items/labels:
        private Font _strongFont = null;
        private Font _regularFont = null;

        private static class SyntaxHiliteGroups
        {
            // Strings sort in order in which regex will be applied:
            public const string VbsComments = "1.1: Comments";
            public const string VbsStrings = "1.2: Strings";
            public const string VbsBrackets = "1.3: Name brackets";
            public const string VbsKeywords = "1.4: VBScript keywords";
            public const string VbsFunctions = "1.5: VBScript functions";

            public const string FilterStrings = "2.1: Strings";
            public const string FilterBrackets = "2.2: Name brackets";
            public const string FilterKeywords = "2.3: Keywords";
            public const string FilterFunctions = "2.4: Functions";

            public const string C1rAggregates = "5.1: C1Report aggregates";
            public const string C1rFunctions = "5.2: C1Report functions"; // Like(), In()
            public const string C1rVariables = "5.3: C1Report variables";

            public const string DbFields = "6.1: Database fields";
        }

        /// <summary>
        /// Specifies the type of syntax highlighting.
        /// </summary>
        public enum SyntaxHighlightingType
        {
            /// <summary>
            /// Syntax highlighting is off.
            /// </summary>
            None,
            /// <summary>
            /// VBScript syntax.
            /// </summary>
            VBScript,
            /// <summary>
            /// MS DataColumn.Expression syntax (https://msdn.microsoft.com/en-us/library/system.data.datacolumn.expression%28v=vs.110%29.aspx).
            /// </summary>
            DataFilter,
        }

        public ScriptEditorControl()
        {
            InitializeComponent();

            _regularFont = inpPanel.Font;
            _strongFont = new System.Drawing.Font(_regularFont, FontStyle.Bold);

        }

        public void Init(C1Report report, EditorScriptItemCollection scriptItems)
        {
            // save parameters
            _report = report;
            _scriptItems = scriptItems;

            AutoScaleMode = AutoScaleMode.None;

            UpdateLists();

            _txtScript.TextChanged += _txtScript_TextChanged;
            _txtScript.SelectionChanged += _txtScript_SelectionChanged;

            imScriptSubs.ChangeCommitted += imScriptSubs_ChangeCommitted;
            imScriptExprs.ChangeCommitted += imScriptExprs_ChangeCommitted;
        }

        void imScriptExprs_ChangeCommitted(object sender, EventArgs e)
        {
            if (imScriptExprs.SelectedOption != null)
            {
                SetCurrentScriptItem(imScriptExprs.SelectedOption.Name, SetCurrentScriptItemContext.ItemSelect);
                _txtScript.Focus();
            }
        }

        void imScriptSubs_ChangeCommitted(object sender, EventArgs e)
        {
            if (imScriptSubs.SelectedOption != null)
            {
                SetCurrentScriptItem(imScriptSubs.SelectedOption.Name, SetCurrentScriptItemContext.ItemSelect);
                _txtScript.Focus();
            }
        }

        void _txtScript_TextChanged(object sender, EventArgs e)
        {
            if (!_currentScriptChangeInternal)
            {
                _currentScriptIsDirty = true;
                if (CurrentScriptTextChanged != null)
                    CurrentScriptTextChanged(this, EventArgs.Empty);
            }
        }

        void _txtScript_SelectionChanged(object sender, EventArgs e)
        {
            if (TextSelectionChanged != null)
                TextSelectionChanged(this, EventArgs.Empty);
        }

        /// <summary>
        /// Fires when a new script is selected by the user, or when a script is loaded initially.
        /// This is NOT the event fired when the text of the script item changes.
        /// </summary>
        public EventHandler CurrentScriptItemChanged;

        /// <summary>
        /// Fires when the user changes (edits) the current script.
        /// </summary>
        public EventHandler CurrentScriptTextChanged;

        public EventHandler TextSelectionChanged;

        /// <summary>
        /// Gets the script editor.
        /// </summary>
        public  RegexTextBox TextBox
        {
            get { return _txtScript; }
        }

        public Font EditorFont
        {
            get { return _txtScript.Font; }
            set
            {
                _currentScriptChangeInternal = true;
                _txtScript.Font = value;
                _currentScriptChangeInternal = false;
            }
        }

        public Color EditorBackgroundColor
        {
            get { return _txtScript.BackColor; }
            set
            {
                _currentScriptChangeInternal = true;
                _txtScript.BackColor = value;
                _currentScriptChangeInternal = false;
            }
        }
        public Color EditorTextColor
        {
            get { return _txtScript.ForeColor; }
            set
            {
                _currentScriptChangeInternal = true;
                _txtScript.ForeColor = value;
                _currentScriptChangeInternal = false;
            }
        }

        public EditorScriptItem CurrentScriptItem
        {
            get
            {
                EditorScriptItem currScriptItem = null;
                if (!string.IsNullOrEmpty(_currentScriptItemKey))
                    _scriptItems.TryGetValue(_currentScriptItemKey, out currScriptItem);
                return currScriptItem;
            }
        }

        public bool IsCurrentScriptDirty
        {
            get { return _currentScriptIsDirty; }
        }

        public bool ShowCalculatedFieldsOnly
        {
            get { return _showCalculatedFieldsOnly; }
            set
            {
                _showCalculatedFieldsOnly = value;
                UpdateLists();
            }
        }

        public bool SyntaxAutoCheck { get; set; }

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public SyntaxHighlightingType SyntaxHighlighting
        {
            get { return _syntaxHighlighting; }
            set
            {
                if (value == _syntaxHighlighting)
                    return;

                _syntaxHighlighting = value;
                switch (_syntaxHighlighting)
                {
                    case SyntaxHighlightingType.None:
                        _txtScript.ClearSyntaxDescriptors();
                        break;
                    case SyntaxHighlightingType.VBScript:
                        _txtScript.ClearSyntaxDescriptors();
                        UpdateSyntaxVBScript(true);
                        UpdateSyntaxC1Report(true);
                        // UpdateSyntaxDataFilter(false);
                        UpdateSyntaxDbFields(true);
                        break;
                    case SyntaxHighlightingType.DataFilter:
                        _txtScript.ClearSyntaxDescriptors();
                        // UpdateSyntaxVBScript(false);
                        // UpdateSyntaxC1Report(false);
                        UpdateSyntaxDataFilter(true);
                        UpdateSyntaxDbFields(true);
                        break;
                    default:
                        System.Diagnostics.Debug.Assert(false);
                        break;
                }
                _txtScript.Refresh();
            }
        }

        public bool AutoSaveCurrentScript { get; set; }

        public enum SetCurrentScriptItemContext
        {
            Initial,
            ItemSelect,
            CloseWindow,
            CancelWindow,
        }

        /// <summary>
        /// Sets the current script item on the control.
        /// If the current item has changed, optionally asks the user to confirm saving the changes.
        /// </summary>
        /// <param name="si"></param>
        /// <param name="confirmUnsavedChanges">If true, and current item has changed, a confirmation message will be shown.
        /// If false, a changed script will be saved automatically.</param>
        /// <returns>True if the change has been successfull, false if the user asked to cancel the change.</returns>
        public bool SetCurrentScriptItem(string scriptItemKey, SetCurrentScriptItemContext context)
        {
            if (_currentScriptItemKey == scriptItemKey)
                return true;

            EditorScriptItem oldScriptItem = null;
            if (!string.IsNullOrEmpty(_currentScriptItemKey))
                _scriptItems.TryGetValue(_currentScriptItemKey, out oldScriptItem);
            EditorScriptItem newScriptItem = null;
            if (!string.IsNullOrEmpty(scriptItemKey))
                _scriptItems.TryGetValue(scriptItemKey, out newScriptItem);

            if (oldScriptItem != null && _currentScriptIsDirty)
            {
                bool showMessage = false;
                string msg = string.Empty;
                switch (context)
                {
                    case SetCurrentScriptItemContext.Initial:
                        showMessage = false;
                        break;
                    case SetCurrentScriptItemContext.ItemSelect:
                        showMessage = !AutoSaveCurrentScript;
                        if (showMessage)
                            msg = Strings.ScriptEditorControl.ConfirmSave + "\r\n\r\n" + Strings.ScriptEditorControl.ConfirmSaveMore;
                        break;
                    case SetCurrentScriptItemContext.CloseWindow:
                        showMessage = false;
                        break;
                    case SetCurrentScriptItemContext.CancelWindow:
                        showMessage = true;
                        msg = Strings.ScriptEditorControl.ConfirmSave;
                        break;
                    default:
                        System.Diagnostics.Debug.Assert(false);
                        break;
                }
                var confirm = showMessage ? 
                    MessageForm.Show(null, MessageForm.Caption.Confirm, msg, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question) : DialogResult.Yes;
                if (confirm == DialogResult.Cancel)
                    return false;
                else if (confirm == DialogResult.Yes)
                {
                    oldScriptItem.WriteValue(_report, _txtScript.Text);

                    InputOption iopt = null;
                    if (imScriptSubs.Items.Contains(oldScriptItem.Key))
                        iopt = (InputOption)imScriptSubs.Items[oldScriptItem.Key];
                    else if (imScriptExprs.Items.Contains(oldScriptItem.Key))
                        iopt = (InputOption)imScriptExprs.Items[oldScriptItem.Key];
                    UpdateInputOption(iopt, oldScriptItem);
                }
            }

            _currentScriptIsDirty = false;
            _currentScriptChangeInternal = true;
            if (newScriptItem != null)
            {
                if (newScriptItem.IsFieldExpression)
                {
                    lblField.Font = _strongFont;
                    lblScript.Font = _regularFont;
                    imScriptExprs.Text = newScriptItem.ToString();
                    imScriptSubs.Text = null;
                }
                else
                {
                    lblScript.Font = _strongFont;
                    lblField.Font = _regularFont;
                    imScriptSubs.Text = newScriptItem.ToString();
                    imScriptExprs.Text = null;
                }
                _txtScript.Text = newScriptItem.Text;
            }
            else
            {
                lblScript.Font = _regularFont;
                lblField.Font = _regularFont;
                imScriptSubs.Text = null;
                imScriptExprs.Text = null;
                _txtScript.Text = string.Empty;
            }
            _txtScript.ClearUndo();
            _currentScriptChangeInternal = false;

            _currentScriptItemKey = scriptItemKey;

            if (CurrentScriptItemChanged != null)
                CurrentScriptItemChanged(this, EventArgs.Empty);

            return true;
        }

        private InputOption MakeInputOption(EditorScriptItem si)
        {
            InputOption iopt = new InputOption();
            iopt.Name = si.Key;
            iopt.Text = si.ToString();
            iopt.Width = imScriptSubs.Width;
            UpdateInputOption(iopt, si);
            return iopt;
        }

        private void UpdateInputOption(InputOption iopt, EditorScriptItem si)
        {
            // iopt.Text = si.ToString();
            if (si.HasText)
            {
                iopt.ToolTipText = string.Format("<pre>{0}</pre>", si.Text);
                if (si.ItemKind != EditorScriptItem.Kind.FieldText)
                    iopt.Font = _strongFont;
                else
                    iopt.Font = _regularFont;
            }
            else
            {
                iopt.ToolTipText = null;
                iopt.Font = _regularFont;
            }
        }


        private void UpdateLists()
        {
            this.inpPanel.BeginUpdate();
            try
            {
                imScriptSubs.Items.Clear();
                imScriptExprs.Items.Clear();

                if (_scriptItems == null)
                    return;

                foreach (string key in _scriptItems.Keys)
                {
                    EditorScriptItem si = _scriptItems[key];
                    if (si.IsFieldExpression)
                    {
                        if (!ShowCalculatedFieldsOnly || si.ItemKind != EditorScriptItem.Kind.FieldText)
                            imScriptExprs.Items.Add(MakeInputOption(si));
                    }
                    else
                        imScriptSubs.Items.Add(MakeInputOption(si));
                }
            }
            finally
            {
                this.inpPanel.EndUpdate();
            }
        }

        public bool CheckSyntax(out string error)
        {
            if (_report != null && SyntaxAutoCheck)
            {
                var si = CurrentScriptItem;
                if (si != null)
                {
                    bool asExpression;
                    if (si.CanSyntaxCheck(out asExpression))
                    {
                        Exception x = asExpression ? _report.ParseExpression(_txtScript.Text) : _report.ParseStatement(_txtScript.Text, si.Key, si.IsGlobal);
                        if (x != null)
                        {
                            error = x.Message;
                            return false;
                        }

                    }
                }
            }
            error = string.Empty;
            return true;
        }

        public void InitSyntaxDbFields(string[] fields)
        {
            _dbFields = fields;
            UpdateSyntaxDbFields(SyntaxHighlighting != SyntaxHighlightingType.None);
            _txtScript.Refresh();
        }

        private void UpdateSyntaxC1Report(bool on)
        {
            System.Diagnostics.Debug.Assert(!on || SyntaxHighlighting == SyntaxHighlightingType.VBScript);

            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.C1rAggregates);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.C1rFunctions);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.C1rVariables);
            if (on && _report != null)
            {
                // Custom functions (aggregates)
                string rx = "Avg|Count|CountDistinct|Min|Max|Range|StDev|StDevP|Sum|Var|VarP";
                rx = @"(?i)(^|\W)(?<match>" + rx + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.C1rAggregates, new SyntaxDescriptor(rx, Color.Blue));

                // Built-in functions (Like(), In())
                rx = @"(?i)(^|\W)(?<match>" + "Like|In" + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.C1rFunctions, new SyntaxDescriptor(rx, Color.Blue));

                // report variables
                StringBuilder sb = new StringBuilder("Page|Pages|Report");
                foreach (Field f in _report.Fields)
                {
                    sb.AppendFormat("|{0}", f.Name);
                }
                foreach (Section s in _report.Sections)
                {
                    sb.AppendFormat("|{0}", s.Name);
                }
                rx = @"(?i)(^|\W)(?<match>" + sb.ToString() + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.C1rVariables, new SyntaxDescriptor(rx, Color.Gray));
            }
        }

        /// <summary>
        /// Adds or removes VBScript-specific syntax.
        /// Does NOT update C1Report aggregates/fields/sections.
        /// </summary>
        private void UpdateSyntaxVBScript(bool on)
        {
            System.Diagnostics.Debug.Assert(!on || SyntaxHighlighting == SyntaxHighlightingType.VBScript);

            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.VbsComments);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.VbsStrings);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.VbsBrackets);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.VbsKeywords);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.VbsFunctions);
            if (on)
            {
                // comments
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.VbsComments, new SyntaxDescriptor("(?m)'.*$", Color.Green));

                // strings
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.VbsStrings, new SyntaxDescriptor(@"""[^""\n]*""", Color.DarkRed));

                // variable names in brackets 
                // (to avoid highlighing vb keywords in the names, e.g. [Start Date])
                // note: use Color.FromArgb(1) instead of black because black is the
                //       default color and may be overwritten by other descriptors.
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.VbsBrackets, new SyntaxDescriptor(@"\[[^\]\n]+\]", Color.FromArgb(1)));

                // VScript keywords
                // make sure longer words come first, e.g. EndIf|End, ElseIf|Else
                string rx = "True|False|Nothing|Null|And|Or|Not|Mod|Const|Dim|Redim|" +
                    "If|Then|ElseIf|Else|EndIf|End|Iif|For|To|Next|Step|While|Wend|Sub|Function";
                rx = @"(?i)(^|\W)(?<match>" + rx + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.VbsKeywords, new SyntaxDescriptor(rx, Color.Blue));

                // VBScript functions
                rx = "Argb|Abs|Acos|Asc|Asin|Atn|CBool|CByte|CCur|CDate|CDbl|Chr|CInt|CLng|Cos|CSng|CStr|" +
                    "Date|DateAdd|DateDiff|DatePart|DateSerial|DateValue|Day|Exp|Fix|Format|" +
                    "FormatCurrency|FormatDateTime|FormatNumber|FormatPercent|Hex|Hour|" +
                    "InputBox|InStr|InStrRev|Int|IsDate|IsEmpty|IsNull|IsNumeric|IsObject|LCase|" +
                    "Left|Len|Log|LTrim|Mid|Minute|Month|MonthName|MsgBox|Now|Oct|Pi|Replace|" +
                    "RGB|Right|Rnd|Round|RTrim|Second|Sgn|Sin|Sign|Space|Sqr|StrComp|String|" +
                    "Tan|Time|Timer|TimeSerial|TimeValue|Trim|TypeName|UCase|VarType|WeekDay|WeekDayName|Year|" +
                    "CBoolInvariant|CByteInvariant|CCurInvariant|CDateInvariant|CDblInvariant|" +
                    "CIntInvariant|CLngInvariant|CSngInvariant|CStrInvariant";
                rx = @"(?i)(^|\W)(?<match>" + rx + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.VbsFunctions, new SyntaxDescriptor(rx, Color.Blue));
            }
        }

        public void UpdateSyntaxDbFields(bool on)
        {
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.DbFields);
            if (on && _dbFields != null)
            {
                // add db fields to syntax
                StringBuilder sb = new StringBuilder();
                foreach (string s in _dbFields)
                    sb.AppendFormat(sb.Length == 0 ? "{0}" : "|{0}", s);

                string rx = @"(?i)(^|\W)(?<match>" + sb.ToString() + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.DbFields, new SyntaxDescriptor(rx, Color.DarkSlateBlue));
            }
        }

        private void UpdateSyntaxDataFilter(bool on)
        {
            System.Diagnostics.Debug.Assert(!on || SyntaxHighlighting == SyntaxHighlightingType.DataFilter);

            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.FilterStrings);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.FilterBrackets);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.FilterKeywords);
            _txtScript.RemoveSyntaxDescriptor(SyntaxHiliteGroups.FilterFunctions);
            if (on)
            {
                // comments - not supported

                // strings
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.FilterStrings, new SyntaxDescriptor(@"'[^'\n]*'", Color.DarkRed));

                // variable names in brackets 
                // (to avoid highlighing vb keywords in the names, e.g. [Start Date])
                // note: use Color.FromArgb(1) instead of black because black is the
                //       default color and may be overwritten by other descriptors.
                // (in filter expressions, back quotes may be used instead of square brackets)
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.FilterBrackets, new SyntaxDescriptor(@"\[[^\]\n]+\]|`[^`\n]+`", Color.FromArgb(1)));

                // keywords
                // make sure longer words come first, e.g. EndIf|End, ElseIf|Else
                string rx = "And|Between|Child|False|In|Is|Like|Not|Null|Or|Parent|True";

                rx = @"(?i)(^|\W)(?<match>" + rx + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.FilterKeywords, new SyntaxDescriptor(rx, Color.Blue));

                // functions
                rx = "Sum|Avg|Min|Max|Count|StDev|Var|" +
                    "CONVERT|LEN|ISNULL|IIF|TRIM|SUBSTRING";
                rx = @"(?i)(^|\W)(?<match>" + rx + @")+(?=$|\W)";
                _txtScript.AddSyntaxDescriptor(SyntaxHiliteGroups.FilterFunctions, new SyntaxDescriptor(rx, Color.Blue));
            }
        }
    }
}
