//----------------------------------------------------------------------------
// ScriptEditorForm.cs
//
// Script Editor form for C1ReportDesigner application
//
// This is a part of the C1ReportDesigner package.
// Copyright (C) GrapeCity, Inc
// All rights reserved.
//
// This source code is intended as a supplement to the
// C1ReportDesigner control and related documentation 
// provided with the product.
// See these sources for detailed information regarding the
// C1ReportDesigner product.
//----------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using System.Text.RegularExpressions;
using System.Threading;

using C1.Win.C1Command;
using C1.C1Preview;
using C1.C1Report;
using C1.Win.C1Ribbon;
using C1.Win.C1Themes;
using ReportDesignerApp.Util;

namespace ReportDesignerApp.Forms
{
    public partial class ScriptEditorFormEx : C1.Win.C1Ribbon.C1RibbonForm
    {
        private MainForm _mainForm = null;
        private C1Report _report; // report that owns the script
        private EditorScriptItemCollection _scriptItems = null;
        // the key for the string that we were called for:
        private string _scriptItemKey = null;
        private const int c_minFontSize = 6;
        private const int c_maxFontSize = 24;
        private List<string> _history = new List<string>();
        private int _historyIdx = -1; // current index in history
        private bool _isMovingInHistory = false;

        #region Constructors
        public ScriptEditorFormEx()
        {
            // initialize components
            InitializeComponent();

        }

        public ScriptEditorFormEx(
            MainForm mainForm,
            C1Report report,
            string script,
            bool isExpression)
            : this()
        {
            AutoScaleMode = AutoScaleMode.None;
            Font = MainForm.DefaultAppFont;
            _mainForm = mainForm;
            _report = report;
            _scriptEditor1.TextBox.Font = _mainForm._scriptEditorFont;
            _scriptEditor1.EditorBackgroundColor = _mainForm._scriptEditorBackColor;
            _scriptEditor1.EditorTextColor = _mainForm._scriptEditorForeColor;
            _scriptEditor1.SyntaxAutoCheck = _mainForm._syntaxCheck;
            _scriptEditor1.AutoSaveCurrentScript = _mainForm._scriptEditorAutoSave;
            _scriptEditor1.ShowCalculatedFieldsOnly = _mainForm._scriptEditorCalculatedFieldsOnly;
            // init options in ribbon UI:
            rbEditorFontName.Text = _mainForm._scriptEditorFont.Name;
            rbBackgroundColor.Color = _mainForm._scriptEditorBackColor;
            rbTextColor.Color = _mainForm._scriptEditorForeColor;
            rbSyntaxHighlight.Checked = _mainForm._syntaxColoring;
            rbSyntaxAutoCheck.Checked = _mainForm._syntaxCheck;
            rbAutoSave.Checked = _mainForm._scriptEditorAutoSave;
            rbCalculatedFieldsOnly.Checked = _mainForm._scriptEditorCalculatedFieldsOnly;
            rbDataSourceFilterWarning.Checked = _mainForm._scriptEditorDataSourceFilterWarning;
            // font size:
            for (int i = c_minFontSize; i <= c_maxFontSize; ++i)
                rbFontSize.Items.Add(i.ToString());
            rbFontSize.Text = ((int)Math.Round(_scriptEditor1.EditorFont.Size)).ToString();

            _scriptEditor1.CurrentScriptItemChanged += (ss, ee) =>
            {
                var si = _scriptEditor1.CurrentScriptItem;

                if (!_isMovingInHistory && si != null)
                {
                    if (_historyIdx < _history.Count - 1)
                        _history.RemoveRange(_historyIdx + 1, _history.Count - _historyIdx - 1);
                    _history.Add(si.Key);
                    _historyIdx = _history.Count - 1;
                }
                
                SetFormCaption(si);
                bool isFilter = si != null && si.ItemKind == EditorScriptItem.Kind.DataSourceFilter;
                bool dummy = false;
                bool canSyntaxCheck = si != null && si.CanSyntaxCheck(out dummy);
                
                // For DataSource.Filter, we disable most of the insertion help:
                _slNote.Visible = isFilter && _mainForm._scriptEditorDataSourceFilterWarning;
                if (isFilter)
                    AutoSizeMessage();
                rmnuReportFields.Enabled = cmnuReportFields.Enabled = rmnuReportGlobals.Enabled = cmnuReportGlobals.Enabled = rmnuReportMore.Enabled = cmnuReportMore.Enabled =
                    rmnuVBScriptFunctions.Enabled = cmnuVBScriptFunc.Enabled = rmnuVBScriptProgram.Enabled = cmnuVBScriptProg.Enabled = !isFilter;
                rbSyntaxAutoCheck.Enabled = canSyntaxCheck;
                SetEditorSyntaxHighlighting(_mainForm._syntaxColoring);
                // Update globals:
                InitInsertUIReportGlobals();
            };
            _scriptEditor1.CurrentScriptTextChanged += (ss, ee) =>
            {
                SetFormCaption(_scriptEditor1.CurrentScriptItem);
                CheckSyntax();
            };

            _scriptEditor1.TextSelectionChanged += (ss, ee) => UpdateStatusBar();

            _scriptItems = new EditorScriptItemCollection(_report);

            // this is patchy but will do for now:
            // - if we are being called from prop grid, then we have EditContext and edit that;
            // - else, if we are called to edit an expression - this is creating a new field;
            // - else, the big "Edit Scripts" button was pressed, and we default to GlobalScripts.
            EditorScriptItem scriptItem;
            if (_report.EditContext == null)
            {
                if (isExpression)
                    scriptItem = EditorScriptItem.FromNewField(_report, script);
                else
                    scriptItem = EditorScriptItem.FromGlobalScripts(_report);
            }
            else
                scriptItem = EditorScriptItem.FromEditContext(_report);

            // A script item may not be "collected" if it is a new calc field being added, or e.g. in Maps there are many expressions:
            if (!_scriptItems.ContainsKey(scriptItem.Key))
                _scriptItems.Add(scriptItem.Key, scriptItem);

            _scriptEditor1.Init(_report, _scriptItems);

            // Set current script item:
            _scriptItemKey = scriptItem.Key;
            System.Diagnostics.Debug.Assert(_scriptItems[_scriptItemKey].Text == script);
            // _scriptItems[si.Key].Text = script;
            _scriptEditor1.SetCurrentScriptItem(_scriptItemKey, Forms.ScriptEditorControl.SetCurrentScriptItemContext.Initial);

            // Init insert UI:
            InitInsertUI();

            // ready to start
            UpdateStatusBar();

            var tl = new C1ThemeLocator(_mainForm.Theme);
            var theme = tl.GetTheme();
            if (theme != null)
            {
                C1ThemeController.ApplyThemeToControlTree(this, theme);
                C1ThemeController.ApplyThemeToObject(contextMenu, theme); //qq - if AlexS does not fix it, save theme and apply it to created menus
            }
        }
        #endregion

        #region Private
        private void SetFormCaption(EditorScriptItem si)
        {
            if (si == null)
                this.Text = Strings.ScriptEditorForm.CaptionEmpty;
            else
            {
                var star = _scriptEditor1.IsCurrentScriptDirty ? '*' : ' ';
                if (si.ItemKind == EditorScriptItem.Kind.DataSourceFilter)
                    this.Text = string.Format(Strings.ScriptEditorForm.CaptionFilterFmt, star);
                else
                    this.Text = string.Format(Strings.ScriptEditorForm.CaptionVBScriptFmt, si.ToString(), star);
            }
        }

        private void SetEditorSyntaxHighlighting(bool value)
        {
            if (value)
            {
                var si = _scriptEditor1.CurrentScriptItem;
                bool isFilter = si != null && si.ItemKind == EditorScriptItem.Kind.DataSourceFilter;
                _scriptEditor1.SyntaxHighlighting = isFilter ? ScriptEditorControl.SyntaxHighlightingType.DataFilter : ScriptEditorControl.SyntaxHighlightingType.VBScript;
            }
            else
                _scriptEditor1.SyntaxHighlighting = ScriptEditorControl.SyntaxHighlightingType.None;
        }

        private void CheckSyntax()
        {
            string error;
            // Note: editor.CheckSyntax only does anything if SyntaxAutoCheck and it can:
            if (_scriptEditor1.CheckSyntax(out error))
            {
                rblText.Text = error;
                rbtnError.ToolTip = error;
                rbtnError.Visible = true;
                ribbonSeparator1.Visible = true;
            }
            else
            {
                rbtnError.Visible = false;
                rblText.Text = string.Empty;
                ribbonSeparator1.Visible = false;
            }
        }

        private void InitInsertUI()
        {
            // C1Report:
            InitInsertUIReportFields();
            // called from onitem changed: InitInsertUIReportGlobals();
            InitInsertUIReportMore();
            // VBScript:
            InitInsertUIVBScriptFunctions();
            InitInsertUIVBScriptProgram();
        }

        private void _addMenus(RibbonMenu rMenu, C1CommandMenu cMenu, string[] texts)
        {
            if (texts == null || texts.Length == 0)
            {
                if (rMenu.Items.Count == 0)
                    rMenu.Enabled = false;
                if (cMenu.CommandLinks.Count == 0)
                    cMenu.Enabled = false;
            }
            else
                foreach (string text in texts)
                {
                    AddRibbonInsertTextButton(rMenu, text);
                    AddCommandMenuItem(cMenu, text);
                }
        }
        
        private void _addMenus(RibbonMenu rMenu, C1CommandMenu cMenu, string texts)
        {
            _addMenus(rMenu, cMenu, texts != null ? texts.Split('|') : null);
        }
        
        private void _addSubMenus(RibbonMenu rParent, C1CommandMenu cParent, string menuName, string[] texts)
        {
            RibbonMenu rMenu = new RibbonMenu();
            rMenu.Text = menuName;
            C1CommandMenu cMenu = new C1CommandMenu();
            cMenu.Text = menuName;
            chMain.Commands.Add(cMenu);
            _addMenus(rMenu, cMenu, texts);
            rParent.Items.Add(rMenu);
            cParent.CommandLinks.Add(new C1CommandLink(cMenu));
        }
        
        private void _addSubMenus(RibbonMenu rParent, C1CommandMenu cParent, string menuName, string texts)
        {
            _addSubMenus(rParent, cParent, menuName, texts != null ? texts.Split('|') : null);
        }

        /// <summary>
        /// Tag indicating that the menu item click should insert its text.
        /// </summary>
        private static object s_rbnInsertTextTag = new object();
        private void AddRibbonInsertTextButton(RibbonMenu menu, string text)
        {
            var btn = new RibbonButton(text);
            btn.Tag = s_rbnInsertTextTag;
            menu.Items.Add(btn);
        }
        
        private void AddCommandMenuItem(C1CommandMenu menu, string text)
        {
            C1Command c = new C1Command();
            c.Text = text;
            c.UserData = s_rbnInsertTextTag;
            chMain.Commands.Add(c);
            menu.CommandLinks.Add(new C1CommandLink(c));
        }

        private void _insertTemplate(string text)
        {
            text = text.Replace(" .. ", "\n\n"); ;

            // insert menu text
            _scriptEditor1.TextBox.SelectedText = text;

            // position cursor
            int openPar = text.IndexOf("(");
            int asterisk = text.IndexOf("*");
            if (asterisk > -1)
            {
                _scriptEditor1.TextBox.SelectionStart -= (text.Length - asterisk);
                _scriptEditor1.TextBox.SelectionLength = 1;
            }
            else if (openPar > -1)
                _scriptEditor1.TextBox.SelectionStart -= (text.Length - openPar - 1);
            _scriptEditor1.TextBox.Focus();
        }

        private void InitInsertUIReportFields()
        {
            string[] names = new string[_report.Fields.Count];
            for (int i = 0; i < _report.Fields.Count; ++i)
                names[i] = _report.Fields[i].Name;
            _addMenus(rmnuReportFields, cmnuReportFields, names);
        }

        private void LoadDBFields()
        {
            lock (_report)
            {
                // get the list of db fields (may be slow the first time)
                _report.DataSource.GetDBFieldList(true);
            }

            // done, do the work on the main thread
            Invoke(new EventHandler(LoadDBFields));
        }

        private void LoadDBFields(object sender, System.EventArgs e)
        {
            // get a list of db fields
            lock (_report)
            {
                string[] dbFields = _report.DataSource.GetDBFieldList(true);
                _addMenus(rmnuDataFields, cmnuDataFields, dbFields);
                // initialize syntax coloring
                _scriptEditor1.InitSyntaxDbFields(dbFields);
            }
        }

        private void InitInsertUIReportGlobals()
        {
            _report.ParseGlobalScript();
            rmnuReportGlobals.Items.Clear();
            rmnuReportGlobals.Enabled = true;
            cmnuReportGlobals.CommandLinks.Clear();
            cmnuReportGlobals.Enabled = true;
            _addMenus(rmnuReportGlobals, cmnuReportGlobals, _report.GetGlobalFunctions());
            _addMenus(rmnuReportGlobals, cmnuReportGlobals, _report.GetGlobalSubs());
        }

        private void InitInsertUIReportMore()
        {
            // Sections:
            string[] names = new string[_report.Sections.Count];
            for (int i = 0; i < _report.Sections.Count; ++i)
                names[i] = _report.Sections[i].Name;
            _addSubMenus(rmnuReportMore, cmnuReportMore, Strings.ScriptEditorForm.MnuReportSections, names);
            // Report variables:
            _addSubMenus(rmnuReportMore, cmnuReportMore, Strings.ScriptEditorForm.MnuReportVariables, "Page|Pages|Report|Cancel");
            // Aggregates:
            _addSubMenus(rmnuReportMore, cmnuReportMore, Strings.ScriptEditorForm.MnuAggregateFunctions, "Avg()|Count()|CountDistinct()|Min()|Max()|Range()|StDev()|StDevP()|Sum()|Var()|VarP()");
            // Built-in functions:
            _addSubMenus(rmnuReportMore, cmnuReportMore, Strings.ScriptEditorForm.MnuBuiltinFunctions, "Like(,)|In(,,,)");
        }

        private void InitInsertUIVBScriptFunctions()
        {
            // String:
            _addSubMenus(rmnuVBScriptFunctions, cmnuVBScriptFunc, Strings.ScriptEditorForm.MnuStringFunctions,
                "Asc()|Chr()|CStr()|Format(,)|FormatCurrency(,)|FormatDateTime(,)|FormatNumber(,)|FormatPercent(,)|Instr(,)|LCase()|Left()|Len()|LTrim()|Mid()|Replace(,)|Right(,)|Space()|Trim()|UCase()|Hex()");
            // Data conversion:
            _addSubMenus(rmnuVBScriptFunctions, cmnuVBScriptFunc, Strings.ScriptEditorForm.MnuDataConversion,
                "CBool()|CByte()|CCur()|CDate()|CDbl()|CInt()|CLng()|CSng()|CStr()|IsEmpty()|TypeName()|VarType()|CBoolInvariant()|CByteInvariant()|CCurInvariant()|CDateInvariant()|CDblInvariant()|CIntInvariant()|CLngInvariant()|CSngInvariant()|CStrInvariant()");
            // Math:
            _addSubMenus(rmnuVBScriptFunctions, cmnuVBScriptFunc, Strings.ScriptEditorForm.MnuMathFunctions,
                "Abs()|Acos()|Asin()|Atn()|Cos()|Exp()|Fix()|Int()|IsNull()|IsNumeric()|IsObject()|Pi()|Oct()|Rnd()|Round()|Sgn()|Sin()|Sign()|Sqr()|Tan()");
            // Date/time:
            _addSubMenus(rmnuVBScriptFunctions, cmnuVBScriptFunc, Strings.ScriptEditorForm.MnuDateTimeFunctions,
                "CDate()|DateAdd()|DateDiff(,)|DatePart(,)|DateSerial()|Day()|Hour()|IsDate()|Minute()|Month()|MonthName()|Now|Second()|Weekday()|WeekdayName()|Year()|Date()|Timer()|TimeSerial()|TimeValue()");
        }

        private void InitInsertUIVBScriptProgram()
        {
            // Program Logic:
            _addSubMenus(rmnuVBScriptProgram, cmnuVBScriptProg, Strings.ScriptEditorForm.MnuProgramLogic,
                "Iif(,,)|If * Then *|If * Then * Else *|If * Then .. Else .. End If|And|Or|Not|True|False");
            /* todo: maybe add those as well?
				// statement keywords
				AddKeyword(new Token("If",				TKID.IF,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Then",			TKID.THEN,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Else",			TKID.ELSE,				TKTYPE.STATEMENT));
				AddKeyword(new Token("ElseIf",			TKID.ELSEIF,			TKTYPE.STATEMENT));
				AddKeyword(new Token("Endif",			TKID.ENDIF,				TKTYPE.STATEMENT));
				AddKeyword(new Token("For",				TKID.FOR,				TKTYPE.STATEMENT));
				AddKeyword(new Token("To",				TKID.TO,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Next",			TKID.NEXT,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Step",			TKID.STEP,				TKTYPE.STATEMENT));
				AddKeyword(new Token("While",			TKID.WHILE,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Wend",			TKID.WEND,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Const",			TKID.CONST,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Dim",				TKID.DIM,				TKTYPE.STATEMENT));
				AddKeyword(new Token("Redim",			TKID.REDIM,				TKTYPE.STATEMENT));
             */
            // Constants:
            _addSubMenus(rmnuVBScriptProgram, cmnuVBScriptProg, Strings.ScriptEditorForm.MnuConstants,
                "vbCr|vbCrLf|vbLf|vbFormFeed|vbNewLine|vbNullChar|vbTab|vbVerticalTab|vbBlack|vbRed|vbGreen|vbYellow|vbBlue|vbMagenta|vbCyan|vbWhite");
            // Misc:
            _addSubMenus(rmnuVBScriptProgram, cmnuVBScriptProg, Strings.ScriptEditorForm.MnuMiscellaneous,
                "InputBox()|MsgBox()|Rgb()|Argb()|Mod|Nothing|Null");
            // Global Script:
            _addSubMenus(rmnuVBScriptProgram, cmnuVBScriptProg, Strings.ScriptEditorForm.MnuFunctionsAndSubs,
                "Function *() .. End Function|Sub *() .. End Sub|As|Call|ByRef|ByVal");
        }

        private void UpdateStatusBar()
        {
            int chr = _scriptEditor1.TextBox.SelectionStart;
            int line = _scriptEditor1.TextBox.GetLineFromCharIndex(chr);

            int column = 0;
            string text = _scriptEditor1.TextBox.Text;
            for (int i = Math.Min(chr - 1, text.Length - 1); i > 0; i--)
            {
                if (text[i] == '\n') break;
                column++;
            }

            rlblCol.Text = string.Format(Strings.ScriptEditorForm.ColumnFmt, column + 1);
            rlblRow.Text = string.Format(Strings.ScriptEditorForm.LineFmt, line + 1);
        }
        #endregion

        #region Protected
        // save dialog bounds to reuse next time
        protected override void OnClosing(CancelEventArgs e)
        {
            if (!_scriptEditor1.SetCurrentScriptItem(null, Forms.ScriptEditorControl.SetCurrentScriptItemContext.CancelWindow))
            {
                e.Cancel = true;
            }
            else
                AppSettings.SaveObject("ScriptEditorForm", new FormState(this));
            base.OnClosing(e);
        }

        // apply dialog bounds (after calling OnLoad)
        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);
            AppSettings.LoadObject("ScriptEditorForm", new FormState(this));
            _scriptEditor1.TextBox.Focus();
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);
            AutoSizeMessage();
        }
        #endregion

        #region Public properties

        public string GetScript()
        {
            return _scriptItems[_scriptItemKey].Text;
        }
        #endregion

        private void ScriptEditorForm_Load(object sender, EventArgs e)
        {
            // go load db fields (use a separate thread since this may take some time)
            Thread t = new Thread(new ThreadStart(LoadDBFields));
            t.Start();

            InitCommands();
        }

        private void InitCommands()
        {
            cPaste.CommandStateQuery += (ss, ee) => ee.Enabled = Clipboard.GetDataObject() != null;
            cCut.CommandStateQuery += (ss, ee) => ee.Enabled = _scriptEditor1.TextBox.SelectionLength > 0;
            cCopy.CommandStateQuery += (ss, ee) => ee.Enabled = _scriptEditor1.TextBox.SelectionLength > 0;
            cDelete.CommandStateQuery += (ss, ee) => ee.Enabled = _scriptEditor1.TextBox.SelectionLength > 0;
            cUndo.CommandStateQuery += (ss, ee) => ee.Enabled = _scriptEditor1.TextBox.CanUndo;
            cRedo.CommandStateQuery += (ss, ee) => ee.Enabled = _scriptEditor1.TextBox.CanRedo;
            cSelectAll.CommandStateQuery += (ss, ee) => ee.Enabled = _scriptEditor1.TextBox.Text.Length > 0;
            cHistBack.CommandStateQuery += (ss, ee) => ee.Enabled = _historyIdx > 0;
            cHistForward.CommandStateQuery += (ss, ee) => ee.Enabled = _historyIdx >= 0 && _historyIdx < _history.Count - 1;
        }

        private void rbnMain_RibbonEvent(object sender, RibbonEventArgs e)
        {
            bool saveAppState = false;
            if (e.EventType == RibbonEventType.Click)
            {
                var button = e.Item as RibbonButton;
                if (button != null && button.Tag == s_rbnInsertTextTag)
                    _insertTemplate(button.Text);
                // the rest is "options":
                else if (button == rbtnFontSizeUp)
                    UpdateEditorFont(new Font(_scriptEditor1.EditorFont.FontFamily, _scriptEditor1.EditorFont.Size + 1));
                else if (button == rbtnFontSizeDown)
                    UpdateEditorFont(new Font(_scriptEditor1.EditorFont.FontFamily, _scriptEditor1.EditorFont.Size - 1));
            }
            else if (e.EventType == RibbonEventType.CheckedChanged)
            {
                if (e.Item ==(object)rbSyntaxHighlight)
                {
                    _mainForm._syntaxColoring = ((RibbonCheckBox)e.Item).Checked;
                    SetEditorSyntaxHighlighting(_mainForm._syntaxColoring);
                    saveAppState = true;
                }
                else if (e.Item == (object)rbSyntaxAutoCheck)
                {
                    _mainForm._syntaxCheck = _scriptEditor1.SyntaxAutoCheck = ((RibbonCheckBox)e.Item).Checked;
                    saveAppState = true;
                }
                else if (e.Item == (object)rbAutoSave)
                {
                    _mainForm._scriptEditorAutoSave = _scriptEditor1.AutoSaveCurrentScript = ((RibbonCheckBox)e.Item).Checked;
                    saveAppState = true;
                }
                else if (e.Item == (object)rbCalculatedFieldsOnly)
                {
                    _mainForm._scriptEditorCalculatedFieldsOnly = _scriptEditor1.ShowCalculatedFieldsOnly = ((RibbonCheckBox)e.Item).Checked;
                    saveAppState = true;
                }
                else if (e.Item == (object)rbDataSourceFilterWarning)
                {
                    _mainForm._scriptEditorDataSourceFilterWarning = ((RibbonCheckBox)e.Item).Checked;
                    saveAppState = true;
                    var si = _scriptEditor1.CurrentScriptItem;
                    if (si != null && si.ItemKind == EditorScriptItem.Kind.DataSourceFilter)
                        _slNote.Visible = _mainForm._scriptEditorDataSourceFilterWarning;
                }
            }
            else if (e.EventType == RibbonEventType.ChangeCommitted)
            {
                if (e.Item == (object)rbEditorFontName)
                    UpdateEditorFont(new Font(((RibbonFontComboBox)e.Item).Text, _scriptEditor1.Font.Size));
                else if (e.Item == (object)rbFontSize)
                    UpdateEditorFont(new Font(_scriptEditor1.EditorFont.FontFamily, int.Parse(rbFontSize.Text)));
            }
            else if (e.EventType == RibbonEventType.SelectedColorChanged)
            {
                if (e.Item == (object)rbBackgroundColor)
                {
                    _mainForm._scriptEditorBackColor = _scriptEditor1.EditorBackgroundColor = ((RibbonColorPicker)e.Item).Color;
                    saveAppState = true;
                }
                else if (e.Item == (object)rbTextColor)
                {
                    _mainForm._scriptEditorForeColor = _scriptEditor1.EditorTextColor = ((RibbonColorPicker)e.Item).Color;
                    saveAppState = true;
                }
            }
            //
            if (saveAppState)
                _mainForm._appConfig.SaveAppState();
        }

        private void UpdateEditorFont(Font font)
        {
            _scriptEditor1.EditorFont = font;
            rbtnFontSizeDown.Enabled = font.Size > c_minFontSize;
            rbtnFontSizeUp.Enabled = font.Size < c_maxFontSize;
            rbEditorFontName.Text = font.Name;
            rbFontSize.Text = font.Size.ToString();
            
            _mainForm._scriptEditorFont = _scriptEditor1.EditorFont;
            _mainForm._appConfig.SaveAppState();
        }

        private void chMain_CommandClick(object sender, CommandClickEventArgs e)
        {
            if (e.Command == cOK)
            {
                _scriptEditor1.SetCurrentScriptItem(null, Forms.ScriptEditorControl.SetCurrentScriptItemContext.CloseWindow);
                this.DialogResult = System.Windows.Forms.DialogResult.OK;
                this.Close();
            }
            else if (e.Command == cCancel)
            {
                // todo - maybe show it anyway? _scriptEditor1.SetCurrentScriptItem(null, true);
                this.DialogResult = System.Windows.Forms.DialogResult.Cancel;
                this.Close();
            }
            else if (e.Command == cHistBack)
            {
                if (_historyIdx > 0)
                {
                    _isMovingInHistory = true;
                    if (_scriptEditor1.SetCurrentScriptItem(_history[_historyIdx - 1], Forms.ScriptEditorControl.SetCurrentScriptItemContext.ItemSelect))
                        --_historyIdx;
                    _isMovingInHistory = false;
                }
            }
            else if (e.Command == cHistForward)
            {
                if (_historyIdx < _history.Count - 1)
                {
                    _isMovingInHistory = true;
                    if (_scriptEditor1.SetCurrentScriptItem(_history[_historyIdx + 1], Forms.ScriptEditorControl.SetCurrentScriptItemContext.ItemSelect))
                        ++_historyIdx;
                    _isMovingInHistory = false;
                }
            }
            else if (e.Command == cUndo)
                _scriptEditor1.TextBox.Undo();
            else if (e.Command == cRedo)
                _scriptEditor1.TextBox.Redo();
            else if (e.Command == cCut)
                _scriptEditor1.TextBox.Cut();
            else if (e.Command == cCopy)
                _scriptEditor1.TextBox.Copy();
            else if (e.Command == cPaste)
                _scriptEditor1.TextBox.Paste();
            else if (e.Command == cDelete)
                _scriptEditor1.TextBox.SelectedText = string.Empty;
            else if (e.Command == cSelectAll)
                _scriptEditor1.TextBox.SelectAll();
            else
            {
                if (e.Command.UserData == s_rbnInsertTextTag)
                    _insertTemplate(e.Command.Text);
            }
        }

        private void rbEditorFontName_ShouldDrawFont(object sender, ShouldDrawFontEventArgs e)
        {
            using (Bitmap b = new Bitmap(1, 1))
            using (Graphics g = Graphics.FromImage(b))
            using (Font f = new Font(e.FontFamily, 10))
                e.ShouldDraw = IsFixedWidth(g, f);
        }

        private bool IsFixedWidth(Graphics g, Font ft)
        {
            char[] charSizes = new char[] { 'i', 'a', 'Z', '%', '#', 'a', 'B', 'l', 'm', ',', '.' };
            float charWidth = g.MeasureString("I", ft).Width;

            foreach (char c in charSizes)
                if (g.MeasureString(c.ToString(), ft).Width != charWidth)
                    return false;

            return true;
        }

        private void _slNote_LinkClicked(object sender, C1.Win.C1SuperTooltip.C1SuperLabelLinkClickedEventArgs e)
        {
            if (e.HRef == "hide_note")
            {
                _slNote.Visible = false;
                _mainForm._scriptEditorDataSourceFilterWarning = rbDataSourceFilterWarning.Checked = false;
                _mainForm._appConfig.SaveAppState();
            }
            else
                MainForm.OpenURL(e.HRef);
        }

        private void AutoSizeMessage()
        {
            // super label's own AutoSize does not seem to work when it is docked, hence this:
            Size s = _slNote.Measure(_slNote.Width);
            _slNote.Height = s.Height;
        }
    }
}
