//----------------------------------------------------------------------------
// AccessFilter.cs
//
// Filter for importing Access reports into the C1ReportDesigner app
// Uses Interop and MSAccess to read and translate report definitions
// stored in mdb and adp files.
//
// This is a part of the C1ReportDesigner package.
// Copyright (C) GrapeCity, Inc
// All rights reserved.
//
// This source code is intended as a supplement to the
// C1ReportDesigner control and related documentation 
// provided with the product.
// See these sources for detailed information regarding the
// C1ReportDesigner product.
//----------------------------------------------------------------------------
using System;
using System.Drawing;
using System.Collections;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using RX = System.Text.RegularExpressions;

using C1.C1Report;
using C1.C1Report.Util;

namespace ReportDesignerApp.Util
{
	/// <summary>
	/// Summary description for AccessFilter.
	/// </summary>
	internal class AccessFilter : IC1ReportImportFilter
	{
		private Form _owner;

		// ** ctor
		public AccessFilter(Form owner)
		{
			_owner = owner;
		}

        private static float ConvertLineSpacing(int lineSpacingTw, Font f)
        {
            float fontHeightTw = f.GetHeight((float)C1.C1Preview.Utils.TwipsPerInch);
            return 100 + lineSpacingTw / fontHeightTw * 100;
        }

		// ** IC1ReportImportFilter
		public event EventHandler Loading;
        bool IC1ReportImportFilter.Import(string mdbFile, ArrayList reportList)
        {
            // always dispose this form when done
            AccessImportForm frm = null;

            // always close Access when done
            Access.Application app = null;

            // assume user wants to clear current list before importing
            bool clearOnImport = true;

            // go to work
            try
            {
                // open mdb file
                app = new Access.ApplicationClass();
                app.OpenCurrentDatabase(mdbFile, false);

                // show import form to select which report to import
                frm = new AccessImportForm();
                foreach (Access.AccessObject obj in app.CurrentProject.AllReports)
                {
                    frm.CheckedList.Items.Add(obj.Name, CheckState.Checked);
                }

                // check that we found at least one report
                if (frm.CheckedList.Items.Count == 0)
                {
                    string msg = string.Format(Strings.AccessFilter.ErrNoReportsFmt, mdbFile);
                    // MessageBox.Show(msg, Strings.AccessFilter.MessageTitle, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    MessageForm.Warn(msg);
                    return false;
                }

                // let user select which reports he wants to import (or cancel the dialog)
                if (frm.ShowDialog() == DialogResult.Cancel)
                {
                    return false;
                }

                // honor 'clear all' setting
                clearOnImport = frm.ClearOnImport;

                // scan again and load all selected reports
				_owner.Cursor = Cursors.WaitCursor;
				foreach (Access.AccessObject obj in app.CurrentProject.AllReports)
                {
                    // check that this one is selected
                    int index = frm.CheckedList.Items.IndexOf(obj.Name);
                    if (index > -1 && frm.CheckedList.GetItemChecked(index) == false)
                        continue;

                    // create report, add to list
                    C1Report c1rpt = new C1Report();
                    c1rpt.ReportName = obj.Name;
                    reportList.Add(c1rpt);

                    // fire event
                    if (Loading != null)
                        Loading(this, EventArgs.Empty);

                    // import this one
                    string name = obj.Name;
                    app.DoCmd.OpenReport(name, Access.AcView.acViewDesign, null, null);
                    TranslateReport(app.Reports[name], c1rpt);
                    app.DoCmd.Close(Access.AcObjectType.acReport, name, Access.AcCloseSave.acSaveNo);
                }

                // load dependent subreports that haven't been selected
                for (int rptIndex = 0; rptIndex < reportList.Count; rptIndex++)
                {
                    C1Report c1rpt = (C1Report)reportList[rptIndex];
                    foreach (Field c1f in c1rpt.Fields)
                    {
                        // look for subreports in this report
                        int pos = c1f.Name.IndexOf('\b');
                        if (pos < 0) continue;
                        string subName = c1f.Name.Substring(pos+1);

                        // if the report hasn't been loaded, load it now
                        int index = frm.CheckedList.Items.IndexOf(subName);
                        if (index > -1 && !frm.CheckedList.GetItemChecked(index))
                        {
                            // add report to list
                            C1Report c1rptSub = new C1Report();
                            c1rptSub.ReportName = subName;
                            reportList.Add(c1rptSub);

                            // import report
                            app.DoCmd.OpenReport(subName, Access.AcView.acViewDesign, null, null);
                            TranslateReport(app.Reports[subName], c1rptSub);
                            app.DoCmd.Close(Access.AcObjectType.acReport, subName, Access.AcCloseSave.acSaveNo);

                            // need to start again
                            rptIndex = 0;
                        }
                    }
                }

                // fix subreport references
                foreach (C1Report c1rpt in reportList)
                {
                    foreach (Field c1f in c1rpt.Fields)
                    {
                        // look for subreports in this report
                        int pos = c1f.Name.IndexOf('\b');
                        if (pos < 0) continue;
                        string subName = c1f.Name.Substring(pos+1);

                        // look for subreport on the list
                        c1f.Name = c1f.Name.Substring(0, pos);
                        foreach (C1Report c1sub in reportList)
                        {
                            if (string.Compare(c1sub.ReportName, subName, true) == 0)
                            {
                                c1f.Subreport = c1sub;
                                break;
                            }
                        }
                    }
                }
            }
            finally // always close Access and dispose form when done
            {
                if (app != null)
                {
                    try
                    {
                        app.CloseCurrentDatabase();
                        app.Quit(Access.AcQuitOption.acQuitSaveNone);
                    }
                    catch { }
                    System.Runtime.InteropServices.Marshal.FinalReleaseComObject(app);
                    app = null;
                }
                if (frm != null)
                {
                    frm.Dispose();
                }
				_owner.Cursor = Cursors.Default;
			}

			// collect to clear Access instance
			System.GC.Collect();
			System.GC.Collect();
			System.GC.Collect();
			
			// done
            return clearOnImport;
        }

        // ** internal/private stuff
        private bool TranslateReport(Access.Report rpt, C1Report c1rpt)
        {
			//------------------------------------------------------------------------------------
			// start fresh
            c1rpt.Clear();

			//------------------------------------------------------------------------------------
			// use Access mode grow/shrink
			c1rpt.GrowShrinkMode = GrowShrinkModeEnum.Access;

			//------------------------------------------------------------------------------------
			// save report name
            c1rpt.ReportName = rpt.Name;

			//------------------------------------------------------------------------------------
			// save report tag, if any
			string tag = rpt.Tag;
			if (tag != null && tag.Length > 0)
				c1rpt.Tag = tag;

            //------------------------------------------------------------------------------------
            // save data source info
			DataSource ds = c1rpt.DataSource;
			ds.ConnectionString = rpt.Application.ADOConnectString;
			ds.RecordSource = TranslateQueryDef(rpt);
			ds.Filter = rpt.Filter;

			//------------------------------------------------------------------------------------
            // save layout
            C1.C1Report.Layout c1l = c1rpt.Layout;
			if (rpt.PrtMip is byte[]) // could be DBNull...
			{
				PRTMIP mip = new PRTMIP((byte[])rpt.PrtMip);
				c1l.Width = mip.xWidth + mip.yColumnSpacing;
				c1l.MarginLeft = mip.xLeftMargin;
				c1l.MarginRight = mip.xRightMargin;
				c1l.MarginTop = mip.yTopMargin;
				c1l.MarginBottom = mip.yBotMargin;
				c1l.Columns = mip.cxColumns;
				if (mip.cxColumns > 1 && mip.rItemLayout == 1953)
					c1l.ColumnLayout = ColumnLayoutEnum.Across;
			}
            c1l.PageHeader = (C1.C1Report.HdrFtrEnum)rpt.PageHeader;
            c1l.PageFooter = (C1.C1Report.HdrFtrEnum)rpt.PageFooter;
            object picData = rpt.PictureData;
            if (picData != null)
            {
                c1l.Picture = TranslateImage(rpt.Application, picData);
                c1l.PictureShow = TranslatePictureShowEnum(rpt.PicturePages);
                c1l.PictureAlign = TranslatePictureAlignEnum(rpt.PictureSizeMode, rpt.PictureAlignment);
            }

            //------------------------------------------------------------------------------------
            // save groups
            for (int i = 0; ; i++)
            {
                // get ith group
                Access.GroupLevel g = null;
                try
                {
                    g = rpt.get_GroupLevel(i);
                }
                catch {}
                if (g == null) break;
                string gname = string.Format("Group{0}", i + 1);
                Group c1g = c1rpt.Groups.Add(gname, string.Empty, SortEnum.NoSort); 

                c1g.GroupBy = TranslateExpression(g.ControlSource);

                // if we have a GroupOn and GroupInterval, apply the modifiers
                short groupOn = g.GroupOn;
                if (groupOn > 0)
                {
                    // get interval
                    int groupInterval = g.GroupInterval;

                    // apply modifier
                    switch (groupOn)
                    {
                        case 1: // Prefix
                            if (groupInterval < 1) groupInterval = 1;
                            c1g.GroupBy = string.Format("Left([{0}],{1})", c1g.GroupBy, groupInterval);
                            break;
                        case 2: // Year
                            c1g.GroupBy = string.Format("Year([{0}])", c1g.GroupBy);
                            break;
                        case 3: // Quarter
                            c1g.GroupBy = string.Format("DatePart(\"q\", [{0}])", c1g.GroupBy);
                            break;
                        case 9: // Interval
                            c1g.GroupBy = string.Format("Int([{0}]/{1})", c1g.GroupBy, groupInterval);
                            break;
                    }
                }

                // set group properties
                c1g.KeepTogether = (KeepTogetherEnum)g.KeepTogether;
                c1g.Sort = g.SortOrder? SortEnum.Descending: SortEnum.Ascending;
            }

            //------------------------------------------------------------------------------------
            // save sections
            int cnt = c1rpt.Sections.Count;
            for (int i = 0; i < cnt; i++) 
            {
                // get ith section
                Access._Section s = null;
                try
                {
                    s = rpt.get_Section(i);
                } 
                catch {}
                Section c1s = c1rpt.Sections[i];
                c1s.Visible = false;
                if (s == null) continue;

                // save section properties
                c1s.Name = s.Name;
                c1s.Height = s.Height;
                try { c1s.Visible = s.Visible; }                 catch {}
                try { c1s.CanGrow = s.CanGrow; }                 catch {}
                try { c1s.CanShrink = s.CanShrink; }             catch {}
                try { c1s.BackColor = ColorTranslator.FromOle(s.BackColor); }       catch {}
                try { c1s.ForcePageBreak = (ForcePageBreakEnum)s.ForceNewPage; }    catch {}
                try { if (s.NewRowOrCol != 0) c1s.ForcePageBreak = ForcePageBreakEnum.Before; } catch {}
                try { c1s.KeepTogether = s.KeepTogether; }       catch {}
                try { c1s.Repeat = s.RepeatSection; }            catch {}

				// save section tag
				tag = s.Tag;
				if (tag != null && tag.Length > 0)
					c1s.Tag = tag;
            }

            //------------------------------------------------------------------------------------
            // save fields
            foreach (Access.Control ctl in rpt.Controls)
            {
                try
                {
                    SaveControl(rpt, c1rpt, ctl);
                }
                catch (Exception x)
                {
                    Debug.WriteLine(x.Message);
                }
            }

            //------------------------------------------------------------------------------------
            // set Section's CanGrow/CanShrink again in case a field messed it up
            cnt = c1rpt.Sections.Count;
            for (int i = 0; i < cnt; i++) 
            {
                try
                {
                    Access._Section s = rpt.get_Section(i);
                    Section c1s = c1rpt.Sections[i];
                    c1s.CanGrow = s.CanGrow;
                    c1s.CanShrink = s.CanShrink;
                }
                catch {}
            }

            //------------------------------------------------------------------------------------
            // done
            return true;
        }
        private void SaveControl(Access.Report rpt, C1Report c1rpt, Access.Control ctl)
        {
            // get control-specific properties
            System.Reflection.PropertyInfo[] pi = ctl.GetType().GetProperties();
            if (pi != null && pi.Length > 0)
            {
                SaveControlReflection(rpt, c1rpt, ctl, pi);
            }
            else
            {
                SaveControlOriginal(rpt, c1rpt, ctl);
            }
        }
        private void SaveControlOriginal(Access.Report rpt, C1Report c1rpt, Access.Control ctl)
        {
            Field c1f = null;
            string tag = null;

            // translate label controls
            if (ctl is Access.Label)
            {
                Access.Label f = ctl as Access.Label;
                //Console.WriteLine("Label " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, f.Caption, Rectangle.Empty);
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;

                c1f.ForeColor = ColorTranslator.FromOle(f.ForeColor);
                c1f.BackColor = ColorTranslator.FromOle(f.BackColor);
                //c1f.BackStyle = (BackStyleEnum)f.BackStyle;
                if (f.BackStyle == 0)
                    c1f.BackColor = Color.Transparent;
                c1f.Font.Name = f.FontName;
                c1f.Font.Size = f.FontSize;
                c1f.Font.Bold = f.FontBold != 0;
                c1f.Font.Italic = f.FontItalic;
                c1f.Font.Underline = f.FontUnderline;
                c1f.Align = TranslateAlignEnum(f.TextAlign);

                c1f.MarginLeft = f.LeftMargin;
                c1f.MarginRight = f.RightMargin;
                c1f.MarginTop = f.TopMargin;
                c1f.MarginBottom = f.BottomMargin;
                c1f.LineSpacing = ConvertLineSpacing(f.LineSpacing, c1f.Font.Font);

                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;

                try { c1f.LinkTarget = f.HyperlinkAddress; }
                catch { }
            }
            else if (ctl is Access.TextBox) // translate textbox controls
            {
                Access.TextBox f = ctl as Access.TextBox;
                //Console.WriteLine("TextBox " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.Text = TranslateExpression(f.ControlSource);
                c1f.RunningSum = (RunningSumEnum)f.RunningSum;
                c1f.Calculated = true;
                c1f.Format = f.Format;
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;
                c1f.CanGrow = f.CanGrow;
                c1f.CanShrink = f.CanShrink;
                c1f.HideDuplicates = f.HideDuplicates;

                c1f.ForeColor = ColorTranslator.FromOle(f.ForeColor);
                c1f.BackColor = ColorTranslator.FromOle(f.BackColor);
                //c1f.BackStyle = (BackStyleEnum)f.BackStyle;
                if (f.BackStyle == 0)
                    c1f.BackColor = Color.Transparent;
                c1f.Font.Name = f.FontName;
                c1f.Font.Size = f.FontSize;
                c1f.Font.Bold = f.FontBold != 0;
                c1f.Font.Italic = f.FontItalic;
                c1f.Font.Underline = f.FontUnderline;
                c1f.Align = TranslateAlignEnum(f.TextAlign);

                c1f.MarginLeft = f.LeftMargin;
                c1f.MarginRight = f.RightMargin;
                c1f.MarginTop = f.TopMargin;
                c1f.MarginBottom = f.BottomMargin;
                c1f.LineSpacing = ConvertLineSpacing(f.LineSpacing, c1f.Font.Font);

                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;
            }
            else if (ctl is Access.Line) // translate line controls
            {
                Access.Line f = ctl as Access.Line;
                //Console.WriteLine("Line " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;

                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;
                c1f.Shape = new LineShape(f.LineSlant ? LineSlantEnum.Up : LineSlantEnum.Down);
            }
            else if (ctl is Access.Rectangle) // translate rectangle controls
            {
                Access.Rectangle f = ctl as Access.Rectangle;
                //Console.WriteLine("Rectangle " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;

                c1f.BackColor = ColorTranslator.FromOle(f.BackColor);
                //c1f.BackStyle = (BackStyleEnum)f.BackStyle;
                if (f.BackStyle == 0)
                    c1f.BackColor = Color.Transparent;

                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;
            }
            else if (ctl is Access.Image) // translate image controls
            {
                Access.Image f = ctl as Access.Image;
                //Console.WriteLine("Image " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;

                c1f.BackColor = ColorTranslator.FromOle(f.BackColor);
                //c1f.BackStyle = (BackStyleEnum)f.BackStyle;
                if (f.BackStyle == 0)
                    c1f.BackColor = Color.Transparent;
                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;

                c1f.Picture = TranslateImage(rpt.Application, f.PictureData);
                c1f.PictureAlign = TranslatePictureAlignEnum(f.SizeMode, f.PictureAlignment);
            }
            else if (ctl is Access.BoundObjectFrame) // translate bound frame controls (e.g. bound images)
            {
                Access.BoundObjectFrame f = ctl as Access.BoundObjectFrame;
                //Console.WriteLine("BoundObjectFrame " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.Picture = f.ControlSource;
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;

                c1f.BackColor = ColorTranslator.FromOle(f.BackColor);
                //c1f.BackStyle = (BackStyleEnum)f.BackStyle;
                if (f.BackStyle == 0)
                    c1f.BackColor = Color.Transparent;
                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;

                c1f.PictureAlign = TranslatePictureAlignEnum(f.SizeMode, 0);
            }
            else if (ctl is Access.ObjectFrame) // object frame controls not supported (e.g. charts)
            {
                Access.ObjectFrame f = ctl as Access.ObjectFrame;
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;

                c1f.BackColor = ColorTranslator.FromOle(f.BackColor);
                //c1f.BackStyle = (BackStyleEnum)f.BackStyle;
                if (f.BackStyle == 0)
                    c1f.BackColor = Color.Transparent;
                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;

                c1f.Align = FieldAlignEnum.CenterMiddle;
                c1f.Text = "** Untranslated ObjectFrame: " + f.Name + " **";
            }
            else if (ctl is Access.CheckBox) // translate checkbox controls
            {
                Access.CheckBox f = ctl as Access.CheckBox;
                //Console.WriteLine("CheckBox " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, f.ControlSource, Rectangle.Empty);
                c1f.Calculated = true;
                c1f.CheckBox = CheckBoxEnum.CheckBox;
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;
            }
            else if (ctl is Access.OptionButton) // translate option button controls
            {
                Access.OptionButton f = ctl as Access.OptionButton;
                //Console.WriteLine("OptionButton " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, f.ControlSource, Rectangle.Empty);
                c1f.Calculated = true;
                c1f.CheckBox = CheckBoxEnum.RadioButton;
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;
            }
            else if (ctl is Access.PageBreak) // translate page breaks
            {
                Access.PageBreak f = ctl as Access.PageBreak;
                //Console.WriteLine("PageBreak " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.ForcePageBreak = ForcePageBreakEnum.Before;
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = 300;
                c1f.Height = 50;
                c1f.Visible = f.Visible;
            }
            else if (ctl is Access.SubForm) // translate subreports
            {
                Access.SubForm f = ctl as Access.SubForm;
                //Console.WriteLine("SubForm " + f.Name);
                c1f = c1rpt.Fields.Add(f.Name, null, Rectangle.Empty);
                c1f.Section = (SectionTypeEnum)f.Section;
                tag = f.Tag;

                c1f.Left = f.Left;
                c1f.Top = f.Top;
                c1f.Width = f.Width;
                c1f.Height = f.Height;
                c1f.Visible = f.Visible;
                c1f.CanGrow = f.CanGrow;
                c1f.CanShrink = f.CanShrink;

                c1f.BorderStyle = (BorderStyleEnum)f.BorderStyle;
                c1f.BorderColor = ColorTranslator.FromOle(f.BorderColor);
                c1f.LineWidth = f.BorderWidth * 20;

                // get subreport name
                string str = f.SourceObject;
                if (str != null && str.Length > 0)
                {
                    // trim report name
                    if (str.StartsWith("Report."))
                        str = str.Substring(7);

                    // append subreport name to field name
                    // (to resolve later, since subreport may not have been loaded yet)
                    // use \b as a marker to find the field later
                    c1f.Name = string.Format("{0}\b{1}", c1f.Name, str);

                    // get subreport child/master fields
                    string strChild = f.LinkChildFields;
                    string strMaster = f.LinkMasterFields;
                    if (strChild != null && strChild.Length > 0 &&
                        strMaster != null && strMaster.Length > 0)
                    {
                        // build subreport filter
                        // this results in an expression such as:
                        // "[ChildField] = '" & [MasterField] & "'"
                        str = string.Empty;
                        string[] strChildArr = strChild.Split(';');
                        string[] strMasterArr = strMaster.Split(';');
                        int cnt = Math.Min(strChildArr.Length, strMasterArr.Length);
                        for (int i = 0; i < cnt; i++)
                        {
                            // extract a child/master pair
                            string strc = TranslateExpression(strChildArr[i]);
                            string strm = TranslateExpression(strMasterArr[i]);

                            // add this child/master pair to filter expression
                            if (str.Length > 0) str += " & \" And \" & ";
                            str += string.Format("\"{0} = '\" & {1} & \"'\"", strc, strm);
                        }

                        // assign subreport filter to field
                        c1f.Text = str;
                        c1f.Calculated = true;
                    }
                }
            }
            else // we don't translate everything...
            {
                Console.WriteLine("*** UNTRANSLATED FIELD: {0}", ctl.GetType().ToString());
            }

            // all fields may have tags
            if (c1f != null && tag != null && tag.Length > 0)
                c1f.Tag = tag;
        }
        private void SaveControlReflection(Access.Report rpt, C1Report c1rpt, Access.Control ctl, System.Reflection.PropertyInfo[] pi)
        {
            // create field, set basic properties
            Field c1f = c1rpt.Fields.Add(null, null, Rectangle.Empty);
            c1f.Name = (string)GetValue(ctl, pi, "Name");
            c1f.Section = (SectionTypeEnum)(short)GetValue(ctl, pi, "Section");
            c1f.Tag = (string)GetValue(ctl, pi, "Tag");
            c1f.Left = (short)GetValue(ctl, pi, "Left");
            c1f.Top = (short)GetValue(ctl, pi, "Top");
            c1f.Visible = (bool)GetValue(ctl, pi, "Visible");

            // set width/height (pagebreak fields don't have this)
            if (GetValue(ctl, pi, "Width") != null)
            {
                c1f.Width = (short)GetValue(ctl, pi, "Width");
                c1f.Height = (short)GetValue(ctl, pi, "Height");
            }
            else
            {
                c1f.Width = 300;
                c1f.Height = 50;
            }

            // set font
            if (GetValue(ctl, pi, "FontName") != null)
            {
                c1f.Font.Name = (string)GetValue(ctl, pi, "FontName");
                c1f.Font.Size = (short)GetValue(ctl, pi, "FontSize");
                c1f.Font.Bold = (short)GetValue(ctl, pi, "FontBold") != 0;
                c1f.Font.Italic = (bool)GetValue(ctl, pi, "FontItalic");
                c1f.Font.Underline = (bool)GetValue(ctl, pi, "FontUnderline");
                c1f.Align = TranslateAlignEnum((byte)GetValue(ctl, pi, "TextAlign"));
            }

            // set margins
            if (GetValue(ctl, pi, "LeftMargin") != null)
            {
                c1f.MarginLeft = (short)GetValue(ctl, pi, "LeftMargin");
                c1f.MarginRight = (short)GetValue(ctl, pi, "RightMargin");
                c1f.MarginTop = (short)GetValue(ctl, pi, "TopMargin");
                c1f.MarginBottom = (short)GetValue(ctl, pi, "BottomMargin");
                c1f.LineSpacing = ConvertLineSpacing((short)GetValue(ctl, pi, "LineSpacing"), c1f.Font.Font);
            }

            // set border
            if (GetValue(ctl, pi, "BorderStyle") != null)
            {
                c1f.BorderStyle = (BorderStyleEnum)(byte)GetValue(ctl, pi, "BorderStyle");
                c1f.BorderColor = ColorTranslator.FromOle((int)GetValue(ctl, pi, "BorderColor"));
                c1f.LineWidth = 20 * (byte)GetValue(ctl, pi, "BorderWidth");
            }

            // set fore/back colors
            if (GetValue(ctl, pi, "ForeColor") != null)
            {
                c1f.ForeColor = ColorTranslator.FromOle((int)GetValue(ctl, pi, "ForeColor"));
            }
            if (GetValue(ctl, pi, "BackColor") != null)
            {
                c1f.BackColor = ColorTranslator.FromOle((int)GetValue(ctl, pi, "BackColor"));
                if ((byte)GetValue(ctl, pi, "BackStyle") == 0)
                {
                    c1f.BackColor = Color.Transparent;
                }
            }

            // set cangrow/canshrink
            if (GetValue(ctl, pi, "CanGrow") != null)
            {
                c1f.CanGrow = (bool)GetValue(ctl, pi, "CanGrow");
                c1f.CanShrink = (bool)GetValue(ctl, pi, "CanShrink");
            }

            // control-specific properties
            string ctlType = ctl.GetType().Name;

            // label
            if (ctlType.IndexOf("Label") > -1)
            {
                c1f.Text = (string)GetValue(ctl, pi, "Caption");
                c1f.Calculated = false;
                //c1f.LinkTarget = (string)GetValue(ctl, pi, "HyperlinkAddress");
                return;
            }

            // textbox
            if (ctlType.IndexOf("TextBox") > -1)
            {
                c1f.Text = TranslateExpression((string)GetValue(ctl, pi, "ControlSource"));
                c1f.Calculated = true;
                c1f.RunningSum = (RunningSumEnum)(byte)GetValue(ctl, pi, "RunningSum");
                object v = GetValue(ctl, pi, "HideDuplicates");
                if (v is bool)
                    c1f.HideDuplicates = (bool)v;
                c1f.Format = (string)GetValue(ctl, pi, "Format");
                return;
            }

            // line
            if (ctlType.IndexOf("Line") > -1)
            {
                //c1f.LineSlant = (bool)GetValue(ctl, pi, "LineSlant") ? LineSlantEnum.Up : LineSlantEnum.Down;
                c1f.Shape = new LineShape((bool)GetValue(ctl, pi, "LineSlant") ? LineSlantEnum.Up : LineSlantEnum.Down);
                return;
            }

            // page break
            if (ctlType.IndexOf("PageBreak") > -1)
            {
                c1f.ForcePageBreak = ForcePageBreakEnum.Before;
                return;
            }

            // rectangle
            if (ctlType.IndexOf("Rectangle") > -1)
            {
                return;
            }

            // unbound image
            if (ctlType.IndexOf("Image") > -1)
            {
                c1f.Picture = TranslateImage(ctl.Application, GetValue(ctl, pi, "PictureData"));
                c1f.PictureAlign = TranslatePictureAlignEnum((byte)GetValue(ctl, pi, "SizeMode"), (byte)GetValue(ctl, pi, "PictureAlignment"));
                return;
            }

            // bound object frame
            if (ctlType.IndexOf("BoundObjectFrame") > -1)
            {
                c1f.Picture = (string)GetValue(ctl, pi, "ControlSource");
                c1f.PictureAlign = TranslatePictureAlignEnum((byte)GetValue(ctl, pi, "SizeMode"), 0);
                return;
            }

            // object frame (not supported)
            if (ctlType.IndexOf("ObjectFrame") > -1)
            {
                c1f.Align = FieldAlignEnum.CenterMiddle;
                c1f.Text = string.Format("** Untranslated ObjectFrame: {0} **", c1f.Name);
                return;
            }

            // check box
            if (ctlType.IndexOf("CheckBox") > -1)
            {
                c1f.Text = (string)GetValue(ctl, pi, "ControlSource");
                c1f.Calculated = true;
                c1f.CheckBox = CheckBoxEnum.CheckBox;
                return;
            }

            // option button
            if (ctlType.IndexOf("OptionButton") > -1)
            {
                c1f.Text = (string)GetValue(ctl, pi, "ControlSource");
                c1f.Calculated = true;
                c1f.CheckBox = CheckBoxEnum.RadioButton;
                return;
            }

            // subreport
            if (ctlType.IndexOf("SubForm") > -1)
            {
                // get subreport name
                string str = (string)GetValue(ctl, pi, "SourceObject");
                if (str != null && str.Length > 0)
                {
                    // trim report name
                    if (str.StartsWith("Report."))
                        str = str.Substring(7);

                    // append subreport name to field name
                    // (to resolve later, since subreport may not have been loaded yet)
                    // use \b as a marker to find the field later
                    c1f.Name = string.Format("{0}\b{1}", c1f.Name, str);

                    // get subreport child/master fields
                    string strChild = (string)GetValue(ctl, pi, "LinkChildFields");
                    string strMaster = (string)GetValue(ctl, pi, "LinkMasterFields");
                    if (strChild != null && strChild.Length > 0 &&
                        strMaster != null && strMaster.Length > 0)
                    {
                        // build subreport filter
                        // this results in an expression such as:
                        // "[ChildField] = '" & [MasterField] & "'"
                        str = string.Empty;
                        string[] strChildArr = strChild.Split(';');
                        string[] strMasterArr = strMaster.Split(';');
                        int cnt = Math.Min(strChildArr.Length, strMasterArr.Length);
                        for (int i = 0; i < cnt; i++)
                        {
                            // extract a child/master pair
                            string strc = TranslateExpression(strChildArr[i]);
                            string strm = TranslateExpression(strMasterArr[i]);

                            // add this child/master pair to filter expression
                            if (str.Length > 0) str += " & \" And \" & ";
                            str += string.Format("\"{0} = '\" & {1} & \"'\"", strc, strm);
                        }

                        // assign subreport filter to field
                        c1f.Text = str;
                        c1f.Calculated = true;
                    }
                }
                return;
            }

            // what type of field is this???
            Console.WriteLine("*** UNTRANSLATED FIELD: {0}", ctl.GetType().ToString());
        }
        private object GetValue(Access.Control ctl, System.Reflection.PropertyInfo[] piArr, string propertyName)
        {
            foreach (System.Reflection.PropertyInfo pi in piArr)
            {
                if (string.Compare(pi.Name, propertyName, true) == 0)
                {
                    return pi.GetValue(ctl, null);
                }
            }
            return null;
        }

        // load bitmap image stored in Access property
        // ** REVIEW: this is a really sneaky/convoluted way to do it, but I couldn't find an
        // easier way to convert the Access binary image data into a real image...
        private Image TranslateImage(Access.Application app, object picData)
        {
            byte[] bytes = picData as byte[];
            if (bytes == null) return null;

            // create image control
            // Set frm = g_msac.CreateForm
            // g_msac.DoCmd.OpenForm frm.Name, acViewDesign, "", "", acFormPropertySettings, acWindowNormal
            //Set img = g_msac.CreateControl(frm.Name, acImage, acDetail)
            Access.Form frm = app.CreateForm(null, null);
            app.DoCmd.OpenForm(frm.Name, Access.AcFormView.acDesign, string.Empty, string.Empty, Access.AcFormOpenDataMode.acFormPropertySettings, Access.AcWindowMode.acWindowNormal, null);
            object imgObject = app.CreateControl(frm.Name, Access.AcControlType.acImage, Access.AcSection.acDetail, null, null, null, null, null, null);
            if (imgObject == null)
            {
                return null;
            }

            // old-style object model
            Access.Image img = imgObject as Access.Image;
            if (img != null)
            {
                // copy image to image control
                img.PictureData = picData;

                // copy new control to the clipboard
                // TODO: review menu indices, try Access97
                app.DoCmd.DoMenuItem(3, 1, 7, null, 70);
                app.DoCmd.DoMenuItem(3, 1, 2, null, 70);
                //g_msac.DoCmd.DoMenuItem acFormDS, acEdit, 6, acMenuVer70        ' Form Design Bar, Edit, Select All
                //g_msac.DoCmd.DoMenuItem acFormDS, acEdit, acCopy, acMenuVer70   ' Form Design Bar, Edit, Copy

                // get image from the clipboard
                IDataObject data = Clipboard.GetDataObject();
                Bitmap bmp = data.GetData(typeof(Bitmap)) as Bitmap;

                // destroy temp form/control
                //g_msac.DoCmd.Close acForm, frm.Name, acSaveNo
                app.DoCmd.Close(Access.AcObjectType.acForm, frm.Name, Access.AcCloseSave.acSaveNo);
                Clipboard.SetDataObject(string.Empty);

                // return what we got
                return bmp;
            }

            // new style (got to use reflection)
            System.Reflection.PropertyInfo pi = imgObject.GetType().GetProperty("PictureData");
            if (pi != null)
            {
                pi.SetValue(imgObject, picData, null);
                app.DoCmd.DoMenuItem(3, 1, 7, null, 70);
                app.DoCmd.DoMenuItem(3, 1, 2, null, 70);
                IDataObject data = Clipboard.GetDataObject();
                Bitmap bmp = data.GetData(typeof(Bitmap)) as Bitmap;
                app.DoCmd.Close(Access.AcObjectType.acForm, frm.Name, Access.AcCloseSave.acSaveNo);
                Clipboard.SetDataObject(string.Empty);
                return bmp;
            }

            // failed...
            return null;
        }

        // translate enumerations
        private FieldAlignEnum TranslateAlignEnum(int l)
        {
            switch (l) 
            {
                case 1: return FieldAlignEnum.LeftTop;
                case 2: return FieldAlignEnum.CenterTop;
                case 3: return FieldAlignEnum.RightTop;
                case 4: return FieldAlignEnum.JustTop;
            }
            return FieldAlignEnum.General;
        }
        private PictureAlignEnum TranslatePictureAlignEnum(int lSizeMode, int lPicAlign)
        {
            switch (lSizeMode) 
            {
                case 1: return PictureAlignEnum.Stretch;
                case 2: return PictureAlignEnum.Zoom;
            }
            switch (lPicAlign) 
            {
                case 0: return PictureAlignEnum.LeftTop;
                case 1: return PictureAlignEnum.RightTop;
                case 2: return PictureAlignEnum.CenterMiddle;
                case 3: return PictureAlignEnum.LeftBottom;
                case 4: return PictureAlignEnum.RightBottom;
            }
            return PictureAlignEnum.Zoom;
        }
        private PictureShowEnum TranslatePictureShowEnum(long l)
        {
            switch (l) 
            {
                case 0: return PictureShowEnum.AllPages;
                case 1: return PictureShowEnum.FirstPage;
                case 2: return PictureShowEnum.NoPages;
            }
            return PictureShowEnum.AllPages;
        }

        // translate RecordSource QueryDefs into SQL
        private string TranslateQueryDef(Access.Report rpt)
        {
			// get raw RecordSource
			string rs = rpt.RecordSource;

			// empty? don't even try it...
			if (rs == null || rs.Length == 0)
				return string.Empty;

			// translate QueryDefs (DAO only, may throw)
			try
			{
				DAO.Database db = rpt.Application.CurrentDb();
				if (db != null)
				{
					DAO.QueryDef qd = db.QueryDefs[rs];
					if (qd != null) rs = qd.SQL;
				}
			}
			catch {}

            // remove any references to forms
            string lower = rs.ToLower();
            if (lower.IndexOf("forms!") > -1 || lower.IndexOf("[forms]") > -1)
            {
                RX.RegexOptions opt = RX.RegexOptions.IgnoreCase | RX.RegexOptions.Multiline;
                rs = RX.Regex.Replace(rs, @"(forms|\[forms\])!\[[^\]]*\]!", string.Empty, opt);
            }

            // done
            return TranslateExpression(rs);
        }

        // translate expressions into VBScript:
        // 1: "* IS NULL"     --> "isnull(*)"
        // 2: "* IS NOT NULL" --> "not isnull(*)"
        // 3: "Year"          --> "TheYear" (to import nwind correctly)
        string TranslateExpression(string str)
        {
            string[] strRegex = 
            {
                // find																// replace
                //----------------------------------------------------------------------------------------
                //(   (expr)   |   var   |  [brk var] )
                "(\\([^()]+\\)|[\\w\\.]+|\\[[^()]+\\])\\s+IS\\s+(NOT\\s+|)NULL",	"$2IsNull($1)",
                "(^|\\W)Year($|\\W)",												"$1TheYear$2",
            };

            // use regexp only when we have to
            string lower = str.ToLower();
            if (lower.IndexOf("null") > -1 || lower.IndexOf("year") > -1)
            {
                RX.RegexOptions opt = RX.RegexOptions.IgnoreCase | RX.RegexOptions.Multiline;
                for (int i = 0; i < strRegex.Length; i += 2)
                    str = RX.Regex.Replace(str, strRegex[i], strRegex[i+1], opt);
            }

            // bracket fields with embedded delimiters
            if (str.IndexOf(' ') > -1 || str.IndexOf('.') > -1)
            {
                char[] delims = "[]()+-*/,=".ToCharArray();
                if (str.IndexOfAny(delims) < 0)
                    str = "[" + str + "]";
            }

            // done
            return str;
        }

        /// <summary>
        /// PRTMIP is the structure used by Access to store page layout information
        /// </summary>
        private struct PRTMIP
        {
            internal int xLeftMargin;
            internal int yTopMargin;
            internal int xRightMargin;
            internal int yBotMargin;
            internal int fDataOnly;
            internal int xWidth;
            internal int yHeight;
            internal int fDefaultSize;
            internal int cxColumns;
            internal int yColumnSpacing;
            internal int xRowSpacing;
            internal int rItemLayout;
            internal int fFastPrint;
            internal int fDatasheet;
            internal PRTMIP(byte[] data)
            {
                MemoryStream ms = new MemoryStream(data);
                BinaryReader br = new BinaryReader(ms);
                xLeftMargin    = br.ReadInt32();
                yTopMargin     = br.ReadInt32();
                xRightMargin   = br.ReadInt32();
                yBotMargin     = br.ReadInt32();
                fDataOnly      = br.ReadInt32();
                xWidth         = br.ReadInt32();
                yHeight        = br.ReadInt32();
                fDefaultSize   = br.ReadInt32();
                cxColumns      = br.ReadInt32();
                yColumnSpacing = br.ReadInt32();
                xRowSpacing    = br.ReadInt32();
                rItemLayout    = br.ReadInt32();
                fFastPrint     = br.ReadInt32();
                fDatasheet     = br.ReadInt32();
                ms.Close();
            }
        }
    }
}
