//----------------------------------------------------------------------------
// CrystalFilter.cs
//
// Filter for importing Crystal reports into the C1ReportDesigner app
// Uses the CrystalReports library to read and translate report definitions
// stored in rpt files.
//
// This is a part of the C1ReportDesigner package.
// Copyright (C) GrapeCity, Inc
// All rights reserved.
//
// This source code is intended as a supplement to the
// C1ReportDesigner control and related documentation 
// provided with the product.
// See these sources for detailed information regarding the
// C1ReportDesigner product.
//----------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.IO;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Reflection;
using System.Diagnostics;
using System.Xml;
using Microsoft.Win32;

using C1.C1Report;

namespace ReportDesignerApp.Util
{
    /// <summary>
    /// Summary description for CrystalFilter.
    /// </summary>
    internal class CrystalFilter : IC1ReportImportFilter
    {
        private const string CRI2008 = "C1CRI2008.4.exe";
        private const string CRI2010 = "C1CRI2010.4.exe";

        private string _crystalFileName;
        private ArrayList _rptList;
        private ImportMessages _messages = new ImportMessages();

        #region Private
        private string SearchInRegistrySubKeys(RegistryKey key, string fileName)
        {
            if (key == null)
                return null;

            //
            string[] subKeys = key.GetSubKeyNames();
            foreach (string subKey in subKeys)
            {
                string skey = subKey.ToLower();
                if (skey.Contains("c1") || skey.Contains("componentone"))
                {
                    RegistryKey sk = key.OpenSubKey(subKey);
                    if (sk != null)
                    {
                        string s = sk.GetValue(null) as string;
                        if (!string.IsNullOrEmpty(s))
                        {
                            string result = Path.Combine(Path.Combine(s, @"..\..\Apps\v4.0"), fileName);
                            if (File.Exists(result))
                                return result;
                            result = Path.Combine(Path.Combine(s, @"..\..\Apps\v2.0"), fileName);
                            if (File.Exists(result))
                                return result;
                            result = Path.Combine(Path.Combine(s, @"..\..\..\Apps\v4.0"), fileName);
                            if (File.Exists(result))
                                return result;
                            result = Path.Combine(Path.Combine(s, @"..\..\..\Apps\v2.0"), fileName);
                            if (File.Exists(result))
                                return result;
                        }
                    }
                }
            }
            return null;
        }

        private string SearchImportExe(string fileName)
        {
            string result = Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), fileName);
            if (File.Exists(result))
                return result;

            // get C1 studio install directory
            result = SearchInRegistrySubKeys(Registry.LocalMachine.OpenSubKey("Software\\Microsoft\\.NETFrameWork\\AssemblyFolders"), fileName);
            if (!string.IsNullOrEmpty(result))
                return result;

            //
            RegistryKey key = Registry.LocalMachine.OpenSubKey("Software\\Microsoft\\.NETFrameWork");
            if (key != null)
            {
                string[] subKeys = key.GetSubKeyNames();
                foreach (string subKey in subKeys)
                {
                    result = SearchInRegistrySubKeys(key.OpenSubKey(subKey + "\\AssemblyFoldersEx"), fileName);
                    if (!string.IsNullOrEmpty(result))
                        return result;
                }
            }

            return null;
        }

        private void ShowInternalError(
            Exception ex)
        {
            MessageForm.Error(string.Format(Strings.CrystalFilter.ErrInternalFmt, ex.Message));
        }

        private bool DoImport(string exeName)
        {
            string reportsFileName = Path.GetTempFileName();
            string messagesFileName = Path.GetTempFileName();
            try
            {
                int errorCode = CrystalImportExeErrorCodes.NoSuchError;

                ProcessStartInfo psi = new ProcessStartInfo(exeName);
                psi.UseShellExecute = false;
                psi.CreateNoWindow = true;

                string args = string.Format("\"{0}\" \"{1}\" \"{2}\"", _crystalFileName, reportsFileName, messagesFileName);

                if (!MessageForm.GetMessageSuppressed(MessageForm.MessageKey.WarnImportCrystalRASNotAvailable))
                {
                    psi.Arguments = "/RAS " + args;
                    Process process = Process.Start(psi);
                    process.WaitForExit();
                    errorCode = process.ExitCode;
                    process.Dispose();
                    if (errorCode == CrystalImportExeErrorCodes.RASNotAvailable)
                    {
                        DialogResult dr = MessageForm.Show(
                            null,
                            MessageForm.Caption.Warning,
                            Strings.CrystalFilter.RASNotAvailable,
                            MessageBoxButtons.YesNo,
                            MessageBoxIcon.Warning,
                            MessageForm.MessageKey.WarnImportCrystalRASNotAvailable);
                        // if user clicked 'no', bail out:
                        if (dr != DialogResult.Yes)
                            return false;
                        // reset error to try import again w/out mandatory RAS:
                        errorCode = CrystalImportExeErrorCodes.NoSuchError;
                    }
                }

                if (errorCode == CrystalImportExeErrorCodes.NoSuchError)
                {
                    psi.Arguments = args;
                    Process process = Process.Start(psi);
                    process.WaitForExit();
                    errorCode = process.ExitCode;
                    process.Dispose();
                }

                //
                if (errorCode == CrystalImportExeErrorCodes.NoSuchError || errorCode == CrystalImportExeErrorCodes.BadArguments)
                    // see CRI2010\Program.cs, ErrorCode == 4 means that CrystallReport enviroment
                    // can not be initializated
                    // ErrorCode == 1 means that count of parameters is not 3, we shoult never be here
                    return false;

                //
                if (!File.Exists(reportsFileName) || new FileInfo(reportsFileName).Length == 0)
                    // file with reports does not exist,  or it is empty
                    return false;

                // load reports
                try
                {
                    XmlDocument doc = new XmlDocument();
                    doc.Load(reportsFileName);
                    string[] reportNames = C1Report.GetReportList(doc);
                    if (reportNames == null)
                        return false;
                    foreach (string s in reportNames)
                    {
                        try
                        {
                            C1Report report = new C1Report();
                            report.Load(doc, s);
                            _rptList.Add(report);
                            if (Loading != null)
                                Loading(this, EventArgs.Empty);
                        }
                        catch
                        {
                        }
                    }
                }
                catch (Exception ex)
                {
                    // some internal error occurs show error message and exit
                    ShowInternalError(ex);
                    return false;
                }

                // load messages
                if (errorCode != CrystalImportExeErrorCodes.ErrorSavingMessages && File.Exists(messagesFileName) && new FileInfo(reportsFileName).Length > 0)
                {
                    try
                    {
                        _messages.Load(messagesFileName);
                    }
                    catch
                    {
                    }
                }

                // looks like we have some reports now
                return true;
            }
            finally
            {
                // delete temporary files
                try
                {
                    if (File.Exists(reportsFileName))
                        File.Delete(reportsFileName);
                    if (File.Exists(messagesFileName))
                        File.Delete(messagesFileName);
                }
                catch
                {
                }
            }
        }
        #endregion

        #region IC1ReportImportFilter implementation
        public event EventHandler Loading;

        bool IC1ReportImportFilter.Import(
            string rptFile,
            ArrayList reportList)
        {
            // unfortunately I can not search standard, recommended method to determine
            // version of installed Crystal Report.
            // So we simple try both exe for VS2010 and VS2008...
            _rptList = reportList;
            _crystalFileName = rptFile;

            //
            string cri2010 = SearchImportExe(CRI2010);
            string cri2008 = SearchImportExe(CRI2008);

            //
            if (string.IsNullOrEmpty(cri2010) || !DoImport(cri2010))
                if (string.IsNullOrEmpty(cri2008) || !DoImport(cri2008))
                {
                    if (string.IsNullOrEmpty(cri2010) && string.IsNullOrEmpty(cri2008))
                        MessageForm.Error(string.Format(Strings.CrystalFilter.ErrNoExecutables, CRI2010, CRI2008));
                    else if (string.IsNullOrEmpty(cri2010))
                        MessageForm.Error(string.Format(Strings.CrystalFilter.ErrNoCRI2010, CRI2010, CRI2008));
                    else if (string.IsNullOrEmpty(cri2008))
                        MessageForm.Error(string.Format(Strings.CrystalFilter.ErrNoCRI2008, CRI2010, CRI2008));
                    else
                        MessageForm.Error(Strings.CrystalFilter.ErrNoCrystal);
                    return false;
                }

            // show warnings...
            if (_messages.Count > 0)
            {
                ImportMessagesForm.DoShowMessages(_messages, Strings.CrystalFilter.MessageTitle, Strings.CrystalFilter.ImportErrors);
            }

            // return false (don't erase list)
            return false;
        }
        #endregion
    }
}
